from enum import Enum
from ..kind import kind_helper


class ShapeDispatchKind(str, Enum):
    """
    Shapes Lookup dispatch commands for

    .. cssclass:: ul-list

        * ``3D Shapes``
        * ``.uno:ArrowShapes``
        * ``.uno:BasicShapes``
        * ``.uno:CalloutShapes``
        * ``.uno:FlowChartShapes``
        * ``.uno:StarShapes``
        * ``.uno:SymbolShapes``

    See Also:
        - :py:meth:`.Draw.add_dispatch_shape`
        - :py:meth:`.Lo.dispatch_cmd`
        - `Dispatch commands <https://wiki.documentfoundation.org/Development/DispatchCommands>`_
    """

    ARROW_SHAPES_CHEVRON = "ArrowShapes.chevron"
    """Chevron"""
    ARROW_SHAPES_CIRCULAR_ARROW = "ArrowShapes.circular-arrow"
    """Circular Arrow"""
    ARROW_SHAPES_CORNER_RIGHT_ARROW = "ArrowShapes.corner-right-arrow"
    """Corner Right Arrow"""
    ARROW_SHAPES_DOWN_ARROW = "ArrowShapes.down-arrow"
    """Down Arrow"""
    ARROW_SHAPES_DOWN_ARROW_CALLOUT = "ArrowShapes.down-arrow-callout"
    """Down Arrow Callout"""
    ARROW_SHAPES_LEFT_ARROW = "ArrowShapes.left-arrow"
    """Left Arrow"""
    ARROW_SHAPES_LEFT_ARROW_CALLOUT = "ArrowShapes.left-arrow-callout"
    """Left Arrow Callout"""
    ARROW_SHAPES_LEFT_RIGHT_ARROW = "ArrowShapes.left-right-arrow"
    """Left and Right Arrow"""
    ARROW_SHAPES_LEFT_RIGHT_ARROW_CALLOUT = "ArrowShapes.left-right-arrow-callout"
    """Left and Right Arrow Callout"""
    ARROW_SHAPES_NOTCHED_RIGHT_ARROW = "ArrowShapes.notched-right-arrow"
    """Notched Right Arrow"""
    ARROW_SHAPES_PENTAGON_RIGHT = "ArrowShapes.pentagon-right"
    """Pentagon"""
    ARROW_SHAPES_QUAD_ARROW = "ArrowShapes.quad-arrow"
    """4-way Arrow"""
    ARROW_SHAPES_QUAD_ARROW_CALLOUT = "ArrowShapes.quad-arrow-callout"
    """	4-way Arrow Callout"""
    ARROW_SHAPES_RIGHT_ARROW = "ArrowShapes.right-arrow"
    """Right Arrow"""
    ARROW_SHAPES_RIGHT_ARROW_CALLOUT = "ArrowShapes.right-arrow-callout"
    """Right Arrow Callout"""
    ARROW_SHAPES_S_SHARPED_ARROW = "ArrowShapes.s-sharped-arrow"
    """S-shaped Arrow"""
    ARROW_SHAPES_SPLIT_ARROW = "ArrowShapes.split-arrow"
    """Split Arrow"""
    ARROW_SHAPES_SPLIT_ROUND_ARROW = "ArrowShapes.split-round-arrow"
    """Right or Left Arrow"""
    ARROW_SHAPES_STRIPED_RIGHT_ARROW = "ArrowShapes.striped-right-arrow"
    """Striped Right Arrow"""
    ARROW_SHAPES_UP_ARROW = "ArrowShapes.up-arrow"
    """Up Arrow"""
    ARROW_SHAPES_UP_ARROW_CALLOUT = "ArrowShapes.up-arrow-callout"
    """Up Arrow Callout"""
    ARROW_SHAPES_UP_DOWN_ARROW = "ArrowShapes.up-down-arrow"
    """Up and Down Arrow"""
    ARROW_SHAPES_UP_DOWN_ARROW_CALLOUT = "ArrowShapes.up-down-arrow-callout"
    """Up and Down Arrow Callout"""
    ARROW_SHAPES_UP_RIGHT_ARROW = "ArrowShapes.up-right-arrow"
    """Up and Right Arrow"""
    ARROW_SHAPES_UP_RIGHT_ARROW_CALLOUT = "ArrowShapes.up-right-arrow-callout"
    """Up and Right Arrow Callout"""
    ARROW_SHAPES_UP_RIGHT_DOWN_ARROW = "ArrowShapes.up-right-down-arrow"
    """Up, Right and Down Arrow"""

    BASIC_SHAPES_BLOCK_ARC = "BasicShapes.block-arc"
    """Block Arc"""
    BASIC_SHAPES_CAN = "BasicShapes.can"
    """Cylinder"""
    BASIC_SHAPES_CIRCLE = "BasicShapes.circle"
    """Circle"""
    BASIC_SHAPES_CIRCLE_PIE = "BasicShapes.circle-pie"
    """Circle Pie"""
    BASIC_SHAPES_CROSS = "BasicShapes.cross"
    """Cross"""
    BASIC_SHAPES_CUBE = "BasicShapes.cube"
    """Cube"""
    BASIC_SHAPES_DIAMOND = "BasicShapes.diamond"
    """Diamond"""
    BASIC_SHAPES_ELLIPSE = "BasicShapes.ellipse"
    """Ellipse"""
    BASIC_SHAPES_FRAME = "BasicShapes.frame"
    """Frame"""
    BASIC_SHAPES_HEXAGON = "BasicShapes.hexagon"
    """Hexagon"""
    BASIC_SHAPES_ISOSCELES_TRIANGLE = "BasicShapes.isosceles-triangle"
    """Isosceles Triangle"""
    BASIC_SHAPES_OCTAGON = "BasicShapes.octagon"
    """Octagon"""
    BASIC_SHAPES_PAPER = "BasicShapes.paper"
    """Folded Corner"""
    BASIC_SHAPES_PARALLELOGRAM = "BasicShapes.parallelogram"
    """Parallelogram"""
    BASIC_SHAPES_PENTAGON = "BasicShapes.pentagon"
    """Regular Pentagon	"""
    BASIC_SHAPES_QUADRAT = "BasicShapes.quadrat"
    """Square"""
    BASIC_SHAPES_RECTANGLE = "BasicShapes.rectangle"
    """Rectangle"""
    BASIC_SHAPES_RIGHT_TRIANGLE = "BasicShapes.right-triangle"
    """Right Triangle"""
    BASIC_SHAPES_RING = "BasicShapes.ring"
    """Ring"""
    BASIC_SHAPES_ROUND_QUADRAT = "BasicShapes.round-quadrat"
    """Square, Rounded"""
    BASIC_SHAPES_ROUND_RECTANGLE = "BasicShapes.round-rectangle"
    """Rectangle, Rounded"""
    BASIC_SHAPES_TRAPEZOID = "BasicShapes.trapezoid"
    """Trapezoid"""

    CALLOUT_SHAPES_CLOUD_CALLOUT = "CalloutShapes.cloud-callout"
    """Cloud"""
    CALLOUT_SHAPES_LINE_CALLOUT_1 = "CalloutShapes.line-callout-1"
    """Line Callout 1"""
    CALLOUT_SHAPES_LINE_CALLOUT_2 = "CalloutShapes.line-callout-2"
    """Line Callout 2"""
    CALLOUT_SHAPES_LINE_CALLOUT_3 = "CalloutShapes.line-callout-3"
    """Line Callout 3"""
    CALLOUT_SHAPES_RECTANGULAR_CALLOUT = "CalloutShapes.rectangular-callout"
    """Rectangular Callout"""
    CALLOUT_SHAPES_ROUND_CALLOUT = "CalloutShapes.round-callout"
    """Round Callout"""
    CALLOUT_SHAPES_ROUND_RECTANGULAR_CALLOUT = "CalloutShapes.round-rectangular-callout"
    """Rounded Rectangular Callout"""

    FLOW_CHART_SHAPES_FLOWCHART_ALTERNATE_PROCESS = "FlowChartShapes.flowchart-alternate-process"
    """Flowchart: Alternate Process"""
    FLOW_CHART_SHAPES_FLOWCHART_CARD = "FlowChartShapes.flowchart-card"
    """Flowchart: Card"""
    FLOW_CHART_SHAPES_FLOWCHART_COLLATE = "FlowChartShapes.flowchart-collate"
    """Flowchart: Collate"""
    FLOW_CHART_SHAPES_FLOWCHART_CONNECTOR = "FlowChartShapes.flowchart-connector"
    """Flowchart: Connector"""
    FLOW_CHART_SHAPES_FLOWCHART_DATA = "FlowChartShapes.flowchart-data"
    """Flowchart: Data"""
    FLOW_CHART_SHAPES_FLOWCHART_DECISION = "FlowChartShapes.flowchart-decision"
    """Flowchart: Decision"""
    FLOW_CHART_SHAPES_FLOWCHART_DELAY = "FlowChartShapes.flowchart-delay"
    """Flowchart: Delay"""
    FLOW_CHART_SHAPES_FLOWCHART_DIRECT_ACCESS_STORAGE = "FlowChartShapes.flowchart-direct-access-storage"
    """Flowchart: Direct Access Storage"""
    FLOW_CHART_SHAPES_FLOWCHART_DISPLAY = "FlowChartShapes.flowchart-display"
    """Flowchart: Display"""
    FLOW_CHART_SHAPES_FLOWCHART_DOCUMENT = "FlowChartShapes.flowchart-document"
    """Flowchart: Document"""
    FLOW_CHART_SHAPES_FLOWCHART_EXTRACT = "FlowChartShapes.flowchart-extract"
    """Flowchart: Extract"""
    FLOW_CHART_SHAPES_FLOWCHART_INTERNAL_STORAGE = "FlowChartShapes.flowchart-internal-storage"
    """Flowchart: Internal Storage"""
    FLOW_CHART_SHAPES_FLOWCHART_MAGNETIC_DISK = "FlowChartShapes.flowchart-magnetic-disk"
    """Flowchart: Magnetic Disc"""
    FLOW_CHART_SHAPES_FLOWCHART_MANUAL_INPUT = "FlowChartShapes.flowchart-manual-input"
    """Flowchart: Manual Input"""
    FLOW_CHART_SHAPES_FLOWCHART_MANUAL_OPERATION = "FlowChartShapes.flowchart-manual-operation"
    """Flowchart: Manual Operation"""
    FLOW_CHART_SHAPES_FLOWCHART_MERGE = "FlowChartShapes.flowchart-merge"
    """Flowchart: Merge"""
    FLOW_CHART_SHAPES_FLOWCHART_MULTIDOCUMENT = "FlowChartShapes.flowchart-multidocument"
    """Flowchart: Multi-document"""
    FLOW_CHART_SHAPES_FLOWCHART_OFF_PAGE_CONNECTOR = "FlowChartShapes.flowchart-off-page-connector"
    """Flowchart: Off-page Connector"""
    FLOW_CHART_SHAPES_FLOWCHART_OR = "FlowChartShapes.flowchart-or"
    """Flowchart: Or"""
    FLOW_CHART_SHAPES_FLOWCHART_PREDEFINED_PROCESS = "FlowChartShapes.flowchart-predefined-process"
    """Flowchart: Predefined Process"""
    FLOW_CHART_SHAPES_FLOWCHART_PREPARATION = "FlowChartShapes.flowchart-preparation"
    """Flowchart: Preparation"""
    FLOW_CHART_SHAPES_FLOWCHART_PROCESS = "FlowChartShapes.flowchart-process"
    """Flowchart: Process"""
    FLOW_CHART_SHAPES_FLOWCHART_PUNCHED_TAPE = "FlowChartShapes.flowchart-punched-tape"
    """Flowchart: Punched Tape"""
    FLOW_CHART_SHAPES_FLOWCHART_SEQUENTIAL_ACCESS = "FlowChartShapes.flowchart-sequential-access"
    """Flowchart: Sequential Access"""
    FLOW_CHART_SHAPES_FLOWCHART_SORT = "FlowChartShapes.flowchart-sort"
    """Flowchart: Sort"""
    FLOW_CHART_SHAPES_FLOWCHART_STORED_DATA = "FlowChartShapes.flowchart-stored-data"
    """Flowchart: Stored Data"""
    FLOW_CHART_SHAPES_FLOWCHART_SUMMING_JUNCTION = "FlowChartShapes.flowchart-summing-junction"
    """Flowchart: Summing Junction"""
    FLOW_CHART_SHAPES_FLOWCHART_TERMINATOR = "FlowChartShapes.flowchart-terminator"
    """Flowchart: Terminator"""

    THREE_D_HALF_SPHERE = "HalfSphere"
    """3D Half-Sphere"""
    THREE_D_CONE = "Cone"
    """3D Cone"""
    THREE_D_CYRAMID = "Cyramid"
    """3D Pyramid"""
    THREE_D_CYLINDER = "Cylinder"
    """3D Cylinder"""
    THREE_D_SPHERE = "Sphere"
    """3D Sphere"""
    THREE_D_TORUS = "Torus"
    """3D Torus"""

    STAR_SHAPES_BANG = "StarShapes.bang"
    """Explosion"""
    STAR_SHAPES_CONCAVE_STAR6 = "StarShapes.concave-star6"
    """6-Point Star, Concave"""
    STAR_SHAPES_DOORPLATE = "StarShapes.doorplate"
    """Doorplate"""
    STAR_SHAPES_HORIZONTAL_SCROLL = "StarShapes.horizontal-scroll"
    """Horizontal Scroll"""
    STAR_SHAPES_SIGNET = "StarShapes.signet"
    """Signet"""
    STAR_SHAPES_STAR_12 = "StarShapes.star12"
    """12-Point Star"""
    STAR_SHAPES_STAR_24 = "StarShapes.star24"
    """24-Point Star"""
    STAR_SHAPES_STAR_4 = "StarShapes.star4"
    """4-Point Star"""
    STAR_SHAPES_STAR_5 = "StarShapes.star5"
    """5-Point Star"""
    STAR_SHAPES_STAR_6 = "StarShapes.star6"
    """6-Point Star"""
    STAR_SHAPES_STAR_8 = "StarShapes.star8"
    """8-Point Star"""
    STAR_SHAPES_VERTICAL_SCROLL = "StarShapes.vertical-scroll"
    """Vertical Scroll"""

    SYMBOL_SHAPES_BRACE_PAIR = "SymbolShapes.brace-pair"
    """Double Brace"""
    SYMBOL_SHAPES_BRACKET_PAIR = "SymbolShapes.bracket-pair"
    """Double Bracket"""
    SYMBOL_SHAPES_CLOUD = "SymbolShapes.cloud"
    """Cloud"""
    SYMBOL_SHAPES_DIAMOND_BEVEL = "SymbolShapes.diamond-bevel"
    """Diamond Bevel"""
    SYMBOL_SHAPES_FLOWER = "SymbolShapes.flower"
    """Flower"""
    SYMBOL_SHAPES_FORBIDDEN = "SymbolShapes.forbidden"
    """Prohibited"""
    SYMBOL_SHAPES_HEART = "SymbolShapes.heart"
    """Heart"""
    SYMBOL_SHAPES_LEFT_BRACE = "SymbolShapes.left-brace"
    """Left Brace"""
    SYMBOL_SHAPES_LEFT_BRACKET = "SymbolShapes.left-bracket"
    """Left Bracket"""
    SYMBOL_SHAPES_LIGHTNING = "SymbolShapes.lightning"
    """Lightning Bolt"""
    SYMBOL_SHAPES_MOON = "SymbolShapes.moon"
    """Moon"""
    SYMBOL_SHAPES_OCTAGON_BEVEL = "SymbolShapes.octagon-bevel"
    """Octagon Bevel"""
    SYMBOL_SHAPES_PUZZLE = "SymbolShapes.puzzle"
    """Puzzle"""
    SYMBOL_SHAPES_QUAD_BEVEL = "SymbolShapes.quad-bevel"
    """Square Bevel"""
    SYMBOL_SHAPES_RIGHT_BRACE = "SymbolShapes.right-brace"
    """Right Brace"""
    SYMBOL_SHAPES_RIGHT_BRACKET = "SymbolShapes.right-bracket"
    """Right Bracket"""
    SYMBOL_SHAPES_SMILEY = "SymbolShapes.smiley"
    """Smiley Face"""
    SYMBOL_SHAPES_SUN = "SymbolShapes.sun"
    """Sun"""

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def from_str(s: str) -> "ShapeDispatchKind":
        """
        Gets an ``DrawingBitmapKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``ShapeKind`` instance.

        Returns:
            ShapeKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, ShapeDispatchKind)



# coding: utf-8
# Python conversion of Info.java by Andrew Davison, ad@fivedots.coe.psu.ac.th
# See Also: https://fivedots.coe.psu.ac.th/~ad/jlop/
from __future__ import annotations
from enum import Enum, IntFlag
from pathlib import Path
import mimetypes
from typing import TYPE_CHECKING, Any, Tuple, List, Type, cast, overload, Optional
import uno
from ..events.event_singleton import _Events
from ..events.lo_named_event import LoNamedEvent
from .sys_info import SysInfo

from com.sun.star.awt import XToolkit
from com.sun.star.beans import XHierarchicalPropertySet
from com.sun.star.beans import XPropertySet
from com.sun.star.container import XContentEnumerationAccess
from com.sun.star.container import XNameAccess
from com.sun.star.container import XNameContainer
from com.sun.star.deployment import XPackageInformationProvider
from com.sun.star.document import XDocumentPropertiesSupplier
from com.sun.star.document import XTypeDetection
from com.sun.star.frame import XModule
from com.sun.star.lang import XMultiServiceFactory
from com.sun.star.lang import XServiceInfo
from com.sun.star.lang import XTypeProvider
from com.sun.star.reflection import XIdlReflection
from com.sun.star.style import XStyleFamiliesSupplier
from com.sun.star.util import XChangesBatch

if TYPE_CHECKING:
    from com.sun.star.awt import FontDescriptor
    from com.sun.star.beans import XPropertyContainer
    from com.sun.star.document import XDocumentProperties
    from com.sun.star.reflection import XIdlMethod


from ooo.dyn.beans.property_value import PropertyValue
from ooo.dyn.beans.property_concept import PropertyConceptEnum
from ooo.dyn.beans.the_introspection import theIntrospection

from . import lo as mLo
from . import file_io as mFileIO
from . import props as mProps

from . import date_time_util as mDate
from ..meta.static_meta import StaticProperty, classproperty
from ..exceptions import ex as mEx
from ..events.args.event_args import EventArgs
from .type_var import PathOrStr


class Info(metaclass=StaticProperty):

    REG_MOD_FNM = "registrymodifications.xcu"
    NODE_PRODUCT = "/org.openoffice.Setup/Product"
    NODE_L10N = "/org.openoffice.Setup/L10N"
    NODE_OFFICE = "/org.openoffice.Setup/Office"

    NODE_PATHS = (NODE_PRODUCT, NODE_L10N)
    # MIME_FNM = "mime.types"

    # from https://wiki.openoffice.org/wiki/Documentation/DevGuide/OfficeDev/Properties_of_a_Filter
    class Filter(IntFlag):
        """
        Every filter inside LibreOffice is specified by the properties of this enum.
        """

        IMPORT = 0x00000001
        """This filter is used only for internal purposes and so can be used only in API calls. Users won't see it ever. """
        EXPORT = 0x00000002
        """The filter supports the service com.sun.star.document.ExportFilter. It will be shown in the dialog "File-Export". If the filter also has the "IMPORT" flag set, it will be shown in the dialog "File-Save". This makes sure that a format that a user chooses in the save dialog can be loaded again. The same is not guaranteed for a format chosen in "File-Export". """
        TEMPLATE = 0x00000004
        """Filter denotes a template filter (means, by default all documents opened by it become an "untitled" one) """
        INTERNAL = 0x00000008
        """This filter is used only for internal purposes and so can be used only in API calls. Users won't see it ever. """
        TEMPLATEPATH = 0x00000010
        """Must always be set together with "TEMPLATE" to make this feature flag work; soon becoming deprecated"""
        OWN = 0x00000020
        """The filter is a native Apache OpenOffice format (ODF or otherwise)."""
        ALIEN = 0x00000040
        """The filter may lose some information upon saving. """
        DEFAULT = 0x00000100
        """This is the "best" filter for the document type it works on that is guaranteed not so lose any data on export. By default this filter will be used or suggested for every storing process unless the user has chosen a different default filter in the options dialog."""
        SUPPORTSSELECTION = 0x00000400
        """Filter can export only the selected part of a document. This information enables Apache OpenOffice to enable a corresponding check box in the "File-Export" dialog."""
        NOTINFILEDIALOG = 0x00001000
        """This filter will not be shown in the file dialog's filter list"""
        NOTINCHOOSER = 0x00002000
        """This filter will not be shown in the dialog box for choosing a filter in case Apache OpenOffice was not able to detect one"""
        READONLY = 0x00010000
        """All documents imported by this filter will automatically be in read-only state"""
        PREFERRED = 0x10000000
        """The filter is preferred in case of multiple filters for the same file type exist in the configuration"""
        THIRDPARTYFILTER = 0x00080000
        """
        The filter is a UNO component filter, as opposed to the internal C++ filters.
        This is an artifact that will vanish over time.

        AKA: 3RDPARTYFILTER
        """

    @staticmethod
    def get_fonts() -> Tuple[FontDescriptor, ...]:
        """
        Gets fonts

        Returns:
            Tuple[FontDescriptor, ...]: Font Descriptors
        """
        xtoolkit = mLo.Lo.create_instance_mcf(XToolkit, "com.sun.star.awt.Toolkit")
        device = xtoolkit.createScreenCompatibleDevice(0, 0)
        if device is None:
            mLo.Lo.print("Could not access graphical output device")
            return ()
        return device.getFontDescriptors()

    @classmethod
    def get_font_names(cls) -> List[str]:
        """
        Gets font names

        Returns:
            List[str]: Font names
        """
        fds = cls.get_fonts()
        if fds is None:
            return []

        names_set = set()
        for name in fds:
            names_set.add(name)
        names = list(names_set)
        names.sort()
        return names

    @staticmethod
    def get_font_mono_name() -> str:
        """
        Gets a general font such as ``Courier New`` (windows) or ``Liberation Mono``

        Returns:
            str: Font Name

        See Also:
            `Fonts <https://wiki.documentfoundation.org/Fonts>`_ on Document Foundation’s wiki
        """
        pf = SysInfo.get_platform()
        if pf == SysInfo.PlatformEnum.WINDOWS:
            return "Courier New"
        else:
            return "Liberation Mono"  # Metrically compatible with Courier New

    @staticmethod
    def get_font_general_name() -> str:
        """
        Gets a general font such as ``Times New Roman`` (windows) or ``Liberation Serif``

        Returns:
            str: Font Name

        See Also:
            `Fonts <https://wiki.documentfoundation.org/Fonts>`_ on Document Foundation’s wiki
        """
        pf = SysInfo.get_platform()
        if pf == SysInfo.PlatformEnum.WINDOWS:
            return "Times New Roman"
        else:
            return "Liberation Serif"  # Metrically compatible with Times New Roman

    @classmethod
    def get_reg_mods_path(cls) -> str:
        """
        Get registered modifications path

        Returns:
            str: registered modifications path
        """
        user_cfg_dir = mFileIO.FileIO.url_to_path(cls.get_paths("UserConfig"))
        parent_path = user_cfg_dir.parent
        return str(parent_path / cls.REG_MOD_FNM)

    @overload
    @classmethod
    def get_reg_item_prop(cls, item: str, prop: str) -> str:
        """
        Gets value from 'registrymodifications.xcu'

        Args:
            item (str): item name
            prop (str): property value

        Raises:
            ValueError: if unable to get value

        Returns:
            str: value from 'registrymodifications.xcu'. e.g. "Writer/MailMergeWizard" null, "MailAddress"
        """
        ...

    @overload
    @classmethod
    def get_reg_item_prop(cls, item: str, prop: str, node: str) -> str:
        """
        Gets value from 'registrymodifications.xcu'

        Args:
            item (str): item name
            prop (str): property value
            node (str): node

        Raises:
            ValueError: if unable to get value

        Returns:
            str: value from 'registrymodifications.xcu'. e.g. "Writer/MailMergeWizard" null, "MailAddress"
        """
        ...

    @classmethod
    def get_reg_item_prop(cls, item: str, prop: str, node: Optional[str] = None) -> str:
        """
        Gets value from ``registrymodifications.xcu``

        Args:
            item (str): item name
            prop (str): property value
            node (str): node

        Raises:
            ValueError: if unable to get value

        Returns:
            str: value from ``registrymodifications.xcu``. e.g. ``Writer/MailMergeWizard``, ``None``, ``MailAddress``
        """
        # return value from "registrymodifications.xcu"
        # e.g. "Writer/MailMergeWizard" null, "MailAddress"
        # e.g. "Logging/Settings", "org.openoffice.logging.sdbc.DriverManager", "LogLevel"
        #
        # This xpath doesn't deal with all cases in the XCU file, which sometimes
        # has many node levels between the item and the prop
        if mLo.Lo.is_macro_mode:
            raise mEx.NotSupportedMacroModeError("get_reg_item_prop() is not supported from a macro")
        try:
            from lxml import etree as XML_ETREE
        except ImportError as e:
            raise Exception("get_reg_item_prop() requires lxml python package") from e

        try:
            _xml_parser = XML_ETREE.XMLParser(remove_blank_text=True)
            fnm = cls.get_reg_mods_path()
            tree: XML_ETREE._ElementTree = XML_ETREE.parse(fnm, parser=_xml_parser)

            if node is None:
                xpath = f"//item[@oor:path='/org.openoffice.Office.{item}']/prop[@oor:name='{prop}']"
            else:
                xpath = f"']/prop[@oor:name='{item}']/node[@oor:name='{node}']/prop[@oor:name='{prop}']"
            value = tree.xpath(xpath)
            if value is None:
                raise Exception("Item Property not found")
            else:
                value = str(value).strip()
                if value == "":
                    raise Exception("Item Property is white space (?)")
            return value
        except Exception as e:
            raise ValueError("unable to get value from registrymodifications.xcu") from e

    @overload
    @classmethod
    def get_config(cls, node_str: str) -> object:
        """
        Get config

        Args:
            node_str (str): node string

        Raises:
            ConfigError: if unable to get config

        Returns:
            object: config
        """
        ...

    @overload
    @classmethod
    def get_config(cls, node_str: str, node_path: str) -> object:
        """
        Get config

        Args:
            node_str (str): node string
            node_path (str): node_path

        Raises:
            ConfigError: if unable to get config

        Returns:
            object: config
        """
        ...

    @classmethod
    def get_config(cls, node_str: str, node_path: Optional[str] = None) -> object:
        """
        Get config

        Args:
            node_str (str): node string
            node_path (str): node_path

        Raises:
            ConfigError: if unable to get config

        Returns:
            object: config

        See Also:
            :ref:`ch03`
        """
        try:
            if node_path is None:
                return cls._get_config2(node_str=node_str)
            return cls._get_config1(node_str=node_str, node_path=node_path)
        except Exception as e:
            msg = f"Unable to get configuration for '{node_str}'"
            if node_path is not None:
                msg += f" with path: '{node_path}'"
            raise mEx.ConfigError(msg) from e

    @classmethod
    def _get_config1(cls, node_str: str, node_path: str):
        props = cls.get_config_props(node_path)
        return mProps.Props.get(props, node_str)

    @classmethod
    def _get_config2(cls, node_str: str) -> object:

        for node_path in cls.NODE_PATHS:
            try:
                return cls._get_config1(node_str=node_str, node_path=node_path)
            except mEx.PropertyNotFoundError:
                pass
        raise mEx.ConfigError(f"{node_str} not found in common node paths")

    @staticmethod
    def get_config_props(node_path: str) -> XPropertySet:
        """
        Get config properties

        Args:
            node_path (str): nod path

        Raises:
            PropertyError: if unable to get get property set

        Returns:
            XPropertySet: Property set
        """
        try:
            con_prov = mLo.Lo.create_instance_mcf(
                XMultiServiceFactory, "com.sun.star.configuration.ConfigurationProvider", raise_err=True
            )
            p = mProps.Props.make_props(nodepath=node_path)
            ca = con_prov.createInstanceWithArguments("com.sun.star.configuration.ConfigurationAccess", p)
            ps = mLo.Lo.qi(XPropertySet, ca, True)
            return ps
        except Exception as e:
            raise mEx.PropertyError(node_path, f"Unable to access config properties for\n\n  '{node_path}'") from e

    @staticmethod
    def get_paths(setting: str) -> str:
        """
        Gets access to LO's predefined paths.

        Args:
            setting (str): property value

        Raises:
            ValueError: if unable to get paths

        Returns:
            str: paths

        Note:
            There are two different groups of properties.
            One group stores only a single path and the other group stores two or
            more paths - separated by a semicolon.

            Some setting values (as listed in the OpenOffice docs for PathSettings)

                - Addin
                - AutoCorrect
                - AutoText
                - Backup
                - Basic
                - Bitmap
                - Config
                - Dictionary
                - Favorite
                - Filter
                - Gallery
                - Graphic
                - Help
                - Linguistic
                - Module
                - Palette
                - Plugin
                - Storage
                - Temp
                - Template
                - UIConfig
                - UserConfig
                - UserDictionary (deprecated)
                - Work

        See Also:
            - :py:meth:`Info.get_dirs`
            - :ref:`ch03`
            - `Wiki Path Settings <https://wiki.openoffice.org/w/index.php?title=Documentation/DevGuide/OfficeDev/Path_Settings>`_
        """
        # access LO's predefined paths. There are two different groups of properties.
        #  One group stores only a single path and the other group stores two or
        #  more paths - separated by a semicolon. See
        #  https://wiki.openoffice.org/w/index.php?title=Documentation/DevGuide/OfficeDev/Path_Settings

        #  Some setting values (as listed in the OpenOffice docs for PathSettings):
        #    Addin, AutoCorrect, AutoText, Backup, Basic, Bitmap,
        #    Config, Dictionary, Favorite, Filter, Gallery,
        #    Graphic, Help, Linguistic, Module, Palette, Plugin,
        #    Storage, Temp, Template, UIConfig, UserConfig,
        #    UserDictionary (deprecated), Work

        # Replaced by thePathSetting in LibreOffice 4.3
        try:
            prop_set = mLo.Lo.create_instance_mcf(XPropertySet, "com.sun.star.util.PathSettings", raise_err=True)
            result = prop_set.getPropertyValue(setting)
            if result is None:
                raise ValueError(f"getPropertyValue() for {setting} yielded None")
            return str(result)
        except Exception as e:
            raise ValueError(f"Could not find paths for: {setting}") from e

    @classmethod
    def get_dirs(cls, setting: str) -> List[Path]:
        """
        Gets dirs paths from settings

        Args:
            setting (str): setting

        Returns:
            List[str]: List of paths

        See Also:
            :py:meth:`~.Info.get_paths`

            `Wiki Path Settings <https://wiki.openoffice.org/w/index.php?title=Documentation/DevGuide/OfficeDev/Path_Settings>`_
        """
        try:
            paths = cls.get_paths(setting)
        except ValueError:
            mLo.Lo.print(f"Cound not find paths for '{setting}'")
            return []
        paths_arr = paths.split(";")
        if len(paths_arr) == 0:
            mLo.Lo.print(f"Cound not split paths for '{setting}'")
            return [mFileIO.FileIO.uri_to_path(paths)]
        dirs = []
        for el in paths_arr:
            dirs.append(mFileIO.FileIO.uri_to_path(el))
        return dirs

    @classmethod
    def get_office_dir(cls) -> str:
        """
        Gets file path to the office dir.
        e.g. ``"C:\Program Files (x86)\LibreOffice 7"``

        Raises:
            ValueError: if unable to obtain office path.

        Returns:
            str: Path as string

        See Also:
            :ref:`ch03`
        """
        try:
            addin_dir = cls.get_paths("Addin")
            # e.g. 'file:///C:/Program%20Files/LibreOffice/program/../program/addin

            addin_path = str(mFileIO.FileIO.uri_to_path(addin_dir))
            #   e.g.  C:\Program%20Files\LibreOffice\program\addin
            try:
                idx = addin_path.index("program")
            except ValueError:
                mLo.Lo.print("Cound not extract office path")
                return addin_path

            p = Path(addin_path[:idx])
            # e.g.  'C:\Program%20Files\LibreOffice\
            return str(p)
        except Exception as e:
            raise ValueError("Unable to get office dir") from e

    @classmethod
    def get_gallery_dir(cls) -> Path:
        """
        Get the first directory that contain the Gallery database and multimedia files.

        Raises:
            ValueError if unable to obtain gallery dir.

        Returns:
            Path: Gallery Dir

        See Also:
            :ref:`ch03`
        """
        try:
            gallery_dirs = cls.get_dirs("Gallery")
            if gallery_dirs is None:
                raise ValueError("No result from get_dir for Gallery")
            return gallery_dirs[0]
        except Exception as e:
            raise ValueError("Unable to get gallery dir") from e

    @classmethod
    def create_configuration_view(cls, path: str) -> XHierarchicalPropertySet:
        """
        Create Configuration View

        Args:
            path (str): path

        Raises:
            ConfigError: if unable to create configuration view

        Returns:
            XHierarchicalPropertySet: Property Set
        """
        try:
            con_prov = mLo.Lo.create_instance_mcf(
                XMultiServiceFactory, "com.sun.star.configuration.ConfigurationProvider"
            )
            if con_prov is None:
                raise mEx.MissingInterfaceError(XMultiServiceFactory)
            _props = mProps.Props.make_props(nodepath=path)
            root = con_prov.createInstanceWithArguments("com.sun.star.configuration.ConfigurationAccess", _props)
            # cls.show_services(obj_name="ConfigurationAccess", obj=root)
            ps = mLo.Lo.qi(XHierarchicalPropertySet, root)
            if ps is None:
                raise mEx.MissingInterfaceError(XHierarchicalPropertySet)
            return ps
        except Exception as e:
            raise mEx.ConfigError(f"Unable to get configuration view for '{path}'") from e

    # =================== update configuration settings ================

    @staticmethod
    def set_config_props(node_path: str) -> XPropertySet:
        """
        Get config properties

        Args:
            node_path (str): Node path of properties

        Raises:
            mEx.ConfigError: if Unable to get config properties

        Returns:
            XPropertySet: Property Set
        """
        try:
            con_prov = mLo.Lo.create_instance_mcf(
                XMultiServiceFactory, "com.sun.star.configuration.ConfigurationProvider"
            )
            if con_prov is None:
                raise mEx.MissingInterfaceError(XMultiServiceFactory)
            _props = mProps.Props.make_props(nodepath=node_path)
            ca = con_prov.createInstanceWithArguments("com.sun.star.configuration.ConfigurationAccess", _props)
            ps = mLo.Lo.qi(XPropertySet, ca)
            if ps is None:
                raise mEx.MissingInterfaceError(XPropertySet)
            return ps
        except Exception as e:
            raise mEx.ConfigError(f"Unable to set configuration property for '{node_path}'") from e

    @classmethod
    def set_config(cls, node_path: str, node_str: str, val: object) -> bool:
        """
        Sets config

        Args:
            node_path (str): node path
            node_str (str): node name
            val (object): node value

        Returns:
            bool: True on success; Otherwise, False
        """
        try:
            props = cls.set_config_props(node_path=node_path)
            if props is None:
                return False
            mProps.Props.set_property(props, node_str, val)
            secure_change = mLo.Lo.qi(XChangesBatch, props)
            if secure_change is None:
                raise mEx.MissingInterfaceError(XChangesBatch)
            secure_change.commitChanges()
            return True
        except Exception as e:
            pass
        return False

    # =================== getting info about a document ====================

    @staticmethod
    def get_name(fnm: PathOrStr) -> str:
        """
        Gets the file's name from the supplied string minus the extension

        Args:
            fnm (PathOrStr): File path

        Raises:
            ValueError: If fnm is empty string
            ValueError: If fnm is not a file

        Returns:
            str: File name minus the extension
        """
        if fnm == "":
            raise ValueError("Empty string")
        p = Path(fnm)
        if not p.is_file():
            raise ValueError(f"Not a file: '{fnm}'")
        if p.suffix == "":
            mLo.Lo.print(f"No extension found for '{fnm}'")
            return p.stem
        return p.stem

    @staticmethod
    def get_ext(fnm: PathOrStr) -> str | None:
        """
        Gets file extension without the ``.``

        Args:
            fnm (PathOrStr): file path

        Raises:
            ValueError: If fnm is empty string

        Returns:
            str | None: Extension if Found; Otherwise, None
        """
        if fnm == "":
            raise ValueError("Empty string")
        p = Path(fnm)
        # if not p.is_file():
        #     mLo.Lo.print(f"Not a file: {fnm}")
        #     return None
        if p.suffix == "":
            mLo.Lo.print(f"No extension found for '{fnm}'")
            return None
        return p.suffix[1:]

    @staticmethod
    def get_unique_fnm(fnm: PathOrStr) -> str:
        """
        If a file called fnm already exists, then a number
        is added to the name so the filename is unique

        Args:
            fnm (str): file path

        Returns:
            str: unique file path
        """
        p = Path(fnm)
        fname = p.stem
        ext = p.suffix
        i = 1
        while p.exists():
            name = f"{fname}{i}{ext}"
            p = p.parent / name
            i += 1
        return str(p)

    @staticmethod
    def get_doc_type(fnm: PathOrStr) -> str:
        """
        Gets doc type from file path

        Args:
            fnm (PathOrStr): File Path

        Raises:
            ValueError: if Unable to get doc type

        Returns:
            str: Doc Type.
        """
        try:
            xdetect = mLo.Lo.create_instance_mcf(XTypeDetection, "com.sun.star.document.TypeDetection", raise_err=True)
            if not mFileIO.FileIO.is_openable(fnm):
                raise mEx.UnOpenableError(fnm)
            url_str = str(mFileIO.FileIO.fnm_to_url(fnm))
            media_desc = (mProps.Props.make_prop_value(name="URL", value=url_str),)

            # even thought queryTypeByDescriptor reports to return a string
            # for some reason I am getting a tuple with the first value as the expected result.
            result = xdetect.queryTypeByDescriptor(media_desc, True)
            if result is None:
                raise mEx.UnKnownError("queryTypeByDescriptor() is an unknown result")
            if isinstance(result, str):
                return result
            return result[0]
        except Exception as e:
            raise ValueError(f"unable to get doc type for '{fnm}'") from e

    @classmethod
    def report_doc_type(cls, doc: object) -> mLo.Lo.DocType:
        """
        Prints doc type to console and return doc type

        Args:
            doc (object): office document

        Returns:
            Lo.DocType: Document type.
        """
        doc_type = mLo.Lo.DocType.UNKNOWN
        if cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.WRITER):
            mLo.Lo.print("A Writer document")
            doc_type = mLo.Lo.DocType.WRITER
        elif cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.IMPRESS):
            mLo.Lo.print("A Impress document")
            doc_type = mLo.Lo.DocType.IMPRESS
        elif cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.DRAW):
            mLo.Lo.print("A Draw document")
            doc_type = mLo.Lo.DocType.DRAW
        elif cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.CALC):
            mLo.Lo.print("A Calc document")
            doc_type = mLo.Lo.DocType.CALC
        elif cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.BASE):
            mLo.Lo.print("A Base document")
            doc_type = mLo.Lo.DocType.BASE
        elif cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.MATH):
            mLo.Lo.print("A Math document")
            doc_type = mLo.Lo.DocType.MATH
        else:
            mLo.Lo.print("Unknown document")
        return doc_type

    @classmethod
    def doc_type_service(cls, doc: object) -> mLo.Lo.Service:
        """
        Prints service type to console and return service type

        Args:
            doc (object): office document

        Returns:
            Lo.Service: Service type
        """
        if cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.WRITER):
            mLo.Lo.print("A Writer document")
            return mLo.Lo.Service.WRITER
        elif cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.IMPRESS):
            mLo.Lo.print("A Impress document")
            return mLo.Lo.Service.IMPRESS
        elif cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.DRAW):
            mLo.Lo.print("A Draw document")
            return mLo.Lo.Service.DRAW
        elif cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.CALC):
            mLo.Lo.print("A Calc document")
            return mLo.Lo.Service.CALC
        elif cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.BASE):
            mLo.Lo.print("A Base document")
            return mLo.Lo.Service.BASE
        elif cls.is_doc_type(obj=doc, doc_type=mLo.Lo.Service.MATH):
            mLo.Lo.print("A Math document")
            return mLo.Lo.Service.MATH
        else:
            mLo.Lo.print("Unknown document")
            return mLo.Lo.Service.UNKNOWN

    @staticmethod
    def is_doc_type(obj: object, doc_type: mLo.Lo.Service) -> bool:
        """
        Gets if doc is a particular doc type.

        Args:
            obj (object): office document
            doc_type (Lo.Service): doc type

        Returns:
            bool: True if obj matches; Otherwise, False
        """
        try:
            si = mLo.Lo.qi(XServiceInfo, obj)
            if si is None:
                return False
            return si.supportsService(str(doc_type))
        except Exception:
            return False

    @staticmethod
    def get_implementation_name(obj: object) -> str:
        """
        Gets implementation name such as ``com.sun.star.comp.deployment.PackageInformationProvider``

        Args:
            obj (object): uno object that implements XServiceInfo

        Raises:
            ValueError: if unable to get implementation name

        Returns:
            str: implementation name
        """
        try:
            si = mLo.Lo.qi(XServiceInfo, obj, True)
            return si.getImplementationName()
        except Exception as e:
            raise ValueError("Could not get service information") from e

    @staticmethod
    def get_identifier(obj: object) -> str:
        """
        Gets identifier name such as ``com.sun.star.text.TextDocument``

        Args:
            obj (object): uno object that implements XModule

        Raises:
            ValueError: if unable to get identifier name

        Returns:
            str: identifier name
        """
        try:
            xmod = mLo.Lo.qi(XModule, obj, True)
            return xmod.getIdentifier()
        except Exception as e:
            raise ValueError("Could not get service information") from e

    @staticmethod
    def get_mime_type(fnm: PathOrStr) -> str:
        """
        Get mime type for a file path

        Args:
            fnm (PathOrStr): file path

        Returns:
            str: Mime type of file if found. Defaults to 'application/octet-stream'
        """
        default = "application/octet-stream"
        mt = mimetypes.guess_type(fnm)
        if mt is None:
            mLo.Lo.print("unable to find mimeypte")
            return default
        if mt[0] is None:
            mLo.Lo.print("unable to find mimeypte")
            return default
        return str(mt[0])

    @staticmethod
    def mime_doc_type(mime_type: str) -> mLo.Lo.DocType:
        """
        Gets document type from mime type

        Args:
            mime_type (str): mime type

        Returns:
            Lo.DocType: Document type. If mime_type is unknown then 'DocType.UNKNOWN'
        """
        if mime_type is None or mime_type == "":
            return mLo.Lo.DocType.UNKNOWN
        if mime_type.find("vnd.oasis.opendocument.text") >= 0:
            return mLo.Lo.DocType.WRITER
        if mime_type.find("vnd.oasis.opendocument.base") >= 0:
            return mLo.Lo.DocType.BASE
        if mime_type.find("vnd.oasis.opendocument.spreadsheet") >= 0:
            return mLo.Lo.DocType.CALC
        if (
            mime_type.find("vnd.oasis.opendocument.graphics") >= 0
            or mime_type.find("vnd.oasis.opendocument.image") >= 0
            or mime_type.find("vnd.oasis.opendocument.chart") >= 0
        ):
            return mLo.Lo.DocType.DRAW
        if mime_type.find("vnd.oasis.opendocument.presentation") >= 0:
            return mLo.Lo.DocType.IMPRESS
        if mime_type.find("vnd.oasis.opendocument.formula") >= 0:
            return mLo.Lo.DocType.MATH
        return mLo.Lo.DocType.UNKNOWN

    @staticmethod
    def is_image_mime(mime_type: str) -> bool:
        """
        Gets if mime-type is a known image type

        Args:
            mime_type (str): mime type e.g. ``application/x-openoffice-bitmap``

        Returns:
            bool: True if known mime-type; Otherwise False
        """
        if mime_type is None or mime_type == "":
            return False
        if mime_type.startswith("image/"):
            return True
        if mime_type.startswith("application/x-openoffice-bitmap"):
            return True
        return False

    # ------------------------ services, interfaces, methods info ----------------------
    @overload
    @classmethod
    def get_service_names(cls) -> List[str]:
        """
        Gets service names

        Returns:
            List[str]: List of service names.
        """
        ...

    @overload
    @classmethod
    def get_service_names(cls, service_name: str) -> List[str]:
        """
        Gets service names

        Args:
            service_name (str): service name

        Returns:
             List[str]: List of service names.
        """
        ...

    @classmethod
    def get_service_names(cls, service_name: Optional[str] = None) -> List[str]:
        """
        Gets service names

        Args:
            service_name (str): service name

        Raises:
            Exception: If error occurs

        Returns:
             List[str]: List of service names.
        """
        if service_name is None:
            return cls._get_service_names1()
        return cls._get_service_names2(service_name=service_name)

    @staticmethod
    def _get_service_names1() -> List[str]:
        mc_factory = mLo.Lo.get_component_factory()
        if mc_factory is None:
            return []
        service_names = list(mc_factory.getAvailableServiceNames())
        service_names.sort()
        return service_names

    @staticmethod
    def _get_service_names2(service_name: str) -> List[str]:
        names: List[str] = []
        try:
            enum_access = mLo.Lo.qi(XContentEnumerationAccess, mLo.Lo.get_component_factory(), True)
            x_enum = enum_access.createContentEnumeration(service_name)
            while x_enum.hasMoreElements():
                si = mLo.Lo.qi(XServiceInfo, x_enum.nextElement())
                names.append(si.getImplementationName())
        except Exception as e:
            mLo.Lo.print(f"Could not collect service names for: {service_name}")
            raise e
        if len(names) == 0:
            mLo.Lo.print(f"No service names found for: {service_name}")
            return names

        names.sort()
        return names

    @staticmethod
    def get_services(obj: object) -> List[str]:
        """
        Gets service names

        Args:
            obj (object): obj that implements XServiceInfo

        Returns:
            List[str]: service names
        """
        try:
            si = mLo.Lo.qi(XServiceInfo, obj, True)
            names = si.getSupportedServiceNames()
            service_names = list(names)
            service_names.sort()
            return service_names
        except Exception as e:
            mLo.Lo.print("Unable to get services")
            mLo.Lo.print(f"    {e}")
            raise e

    @classmethod
    def show_services(cls, obj_name: str, obj: object) -> None:
        """
        Prints services to console

        Args:
            obj_name (str): service name
            obj (object): obj that implements XServiceInfo
        """
        services = cls.get_services(obj=obj)
        if services is None:
            print(f"No supported services found for {obj_name}")
            return
        print(f"{obj_name} Supported Services ({len(services)})")
        for service in services:
            print(f"'{service}'")

    @staticmethod
    def support_service(obj: object, *service: str) -> bool:
        """
        Gets if ``obj`` supports a service.

        Args:
            obj (object): Object to check for supported service
            *service (str): Variable length argument list of UNO namespace strings such as ``com.sun.star.configuration.GroupAccess``

        Returns:
            bool: ``True`` if ``obj`` supports any passed in service; Otherwise, ``False``
        """

        result = False
        try:
            si = mLo.Lo.qi(XServiceInfo, obj)
            if si is None:
                return result
            for srv in service:
                result = si.supportsService(srv)
                if result:
                    break
        except Exception as e:
            mLo.Lo.print("Errors ocurred in support_service(). Returning False")
            mLo.Lo.print(f"    {e}")
        return result

    @staticmethod
    def get_available_services(obj: object) -> List[str]:
        """
        Gets available services for obj

        Args:
            obj (object): obj that implements XMultiServiceFactory interface

        Raises:
            Exception: If unable to get services

        Returns:
            List[str]: List of services
        """
        services: List[str] = []
        try:
            sf = mLo.Lo.qi(XMultiServiceFactory, obj, True)
            service_names = sf.getAvailableServiceNames()
            services.extend(service_names)
            services.sort()
        except Exception as e:
            mLo.Lo.print(e)
            raise Exception() from e
        return services

    @staticmethod
    def get_interface_types(target: object) -> Tuple[object, ...]:
        """
        Get interface types

        Args:
            target (object): object that implements XTypeProvider interface

        Raises:
            Exception: If unable to get services

        Returns:
            Tuple[object, ...]: Tuple of interfaces
        """
        try:
            tp = mLo.Lo.qi(XTypeProvider, target, True)
            types = tp.getTypes()
            return types
        except Exception as e:
            mLo.Lo.print("Unable to get interface types")
            mLo.Lo.print(f"    {e}")
            raise Exception() from e

    @overload
    @classmethod
    def get_interfaces(cls, target: object) -> List[str]:
        """
        Gets interfaces

        Args:
            target: (object): object that implements XTypeProvider

        Returns:
            List[str]: List of interfaces
        """
        ...

    @overload
    @classmethod
    def get_interfaces(cls, type_provider: XTypeProvider) -> List[str]:
        """
        Gets interfaces

        Args:
            type_provider (XTypeProvider): type provider

        Returns:
            List[str]: List of interfaces
        """
        ...

    @classmethod
    def get_interfaces(cls, *args, **kwargs) -> List[str]:
        """
        Gets interfaces

        Args:
            target: (object): object that implements XTypeProvider
            type_provider (XTypeProvider): type provider

        Raises:
            Exception: If unable to get interfaces

        Returns:
            List[str]: List of interfaces
        """
        ordered_keys = (1,)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("target", "typeProvider")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("get_interfaces() got an unexpected keyword argument")
            keys = ("target", "typeProvider")
            for key in keys:
                if key in kwargs:
                    ka[1] = kwargs[key]
                    break
            return ka

        if count != 1:
            raise TypeError("get_interfaces() got an invalid numer of arguments")

        kargs = get_kwargs()

        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg

        try:
            if mLo.Lo.is_uno_interfaces(kargs[1], XTypeProvider):
                type_provider = cast(XTypeProvider, kargs[1])
            else:
                type_provider = mLo.Lo.qi(XTypeProvider, kargs[1])
                if type_provider is None:
                    raise mEx.MissingInterfaceError(XTypeProvider)

            types = cast(Tuple[uno.Type, ...], type_provider.getTypes())
            # use a set to exclude duplicate names
            names_set = set()
            for t in types:
                names_set.add(t.typeName)
            type_names = list(names_set)
            type_names.sort()
            return type_names
        except Exception as e:
            mLo.Lo.print("Unable to get interfaces")
            mLo.Lo.print(f"    {e}")
            raise Exception() from e

    @classmethod
    def is_interface_obj(cls, obj: Any) -> bool:
        """
        Gets is an object contains interfaces

        Args:
            obj (Any): Object to check

        Returns:
            bool: True if obj contains interface; Otherwise, False
        """
        result = False
        if obj is None:
            return result
        try:
            interfaces = cls.get_interfaces(obj)
            result = len(interfaces) > 0
        except mEx.MissingInterfaceError:
            pass
        return result

    @staticmethod
    def is_struct(obj: Any) -> bool:
        """
        Gets if an object is a UNO Struct

        Args:
            obj (Any): Object to check

        Returns:
            bool: True if obj is Struct; Otherwise, False
        """
        if obj is None:
            return False
        if not hasattr(obj, "typeName"):
            return False
        try:
            t = uno.getTypeByName(obj.typeName)
            return t.typeClass.value == "STRUCT"
        except Exception:
            pass
        return False

    @classmethod
    def is_same(cls, obj1: Any, obj2: Any) -> bool:
        """
        Determines if two Uno object are the same.

        Args:
            obj1 (Any): First Uno object
            obj2 (Any): Second Uno Object

        Returns:
            bool: True if objects are the same; Otherwise False.
        """
        # https://wiki.openoffice.org/wiki/Documentation/DevGuide/WritingUNO/XInterface
        # In C++, two objects are the same if their XInterface are the same. The queryInterface() for XInterface will have to
        # be called on both. In Java, check for the identity by calling the runtime function
        # com.sun.star.uni.UnoRuntime.areSame().
        if cls.is_struct(obj1) and cls.is_struct(obj2):
            # types and attribue values must match
            if obj1.typeName != obj2.typeName:
                return False
            obj1_attrs = [s for s in dir(obj1.value) if not s.startswith("_")]
            for atr in obj1_attrs:
                if getattr(obj1, atr) != getattr(obj2, atr):
                    return False
            return True
        elif cls.is_interface_obj(obj1) and cls.is_interface_obj(obj2):
            # must be same object in memory
            id1 = id(obj1)
            id2 = id(obj2)
            return id1 == id2
        return obj1 == obj2

    @staticmethod
    def show_container_names(print_name: str, nc: XNameContainer):
        """
        Prints Name Container elements to console

        Args:
            print_name (str): Name to display. Can be empty string.
            nc (XNameContainer): Name Container to print element names of.
        """
        names = nc.getElementNames()
        if print_name:
            print(f"No. of Names in {print_name}: {len(names)}")
        else:
            print(f"No. of Names in Name Container: {len(names)}")
        for name in names:
            print(f"  {name}")
        print()

    @classmethod
    def show_interfaces(cls, obj_name: str, obj: object) -> None:
        """
        prints interfaces in obj to console

        Args:
            obj_name (str): Name of object for printing
            obj (object): obj that contains interfaces.
        """
        intfs = cls.get_interfaces(obj)
        if not intfs:
            print(f"No interfaces found for {obj_name}")
            return
        print(f"{obj_name} Interfaces ({len(intfs)})")
        for s in intfs:
            print(f"  {s}")

    @staticmethod
    def get_methods_obj(obj: object, property_concept: PropertyConceptEnum | None = None) -> List[str]:
        """
        Get Methods of an object such as a doc.

        Args:
            obj (object): Object to get methods of.
            property_concept (PropertyConceptEnum | None, optional): Type of method to get. Defaults to PropertyConceptEnum.ALL.

        Raises:
            Exception: If unable to get Methods

        Returns:
            List[str]: List of method names found for obj
        """
        if property_concept is None:
            property_concept = PropertyConceptEnum.ALL

        try:
            intro = theIntrospection()
            result = intro.inspect(obj)
            methods = result.getMethods(int(property_concept))
            lst = []
            for meth in methods:
                lst.append(meth.getName())
            lst.sort()
            return lst
        except Exception as e:
            raise Exception(f"Could not get object Methods") from e

    @staticmethod
    def get_methods(interface_name: str) -> List[str]:
        """
        Get Interface Methods

        Args:
            interface_name (str): name of interface

        Returns:
            List[str]: List of methods
        """
        # from com.sun.star.beans.PropertyConcept import ALL
        # ctx = XSCRIPTCONTEXT.getComponentContext()
        # smgr = ctx.getServiceManager()
        # intro = smgr.createInstance('com.sun.star.beans.Introspection')
        # document = XSCRIPTCONTEXT.getDocument()
        # result = intro.inspect(document)
        # methods = result.getMethods(ALL)
        # meth = methods[0]
        #   meth is com.sun.star.reflection.XIdlMethod
        # meth.Name
        #   queryInterface
        #
        #  reflection = smgr.createInstanceWithContext('com.sun.star.reflection.CoreReflection', ctx)
        # See Also: https://tinyurl.com/y338vwhw#L492
        # See Also: https://github.com/hanya/MRI/wiki/RunMRI#Python
        # See Also: https://tinyurl.com/y3m4tx9r#L268

        reflection = mLo.Lo.create_instance_mcf(
            XIdlReflection, "com.sun.star.reflection.CoreReflection", raise_err=True
        )

        fname = reflection.forName(interface_name)  # returns type from name.

        if fname is None:
            mLo.Lo.print(f"Could not find the interface name: {interface_name}")
            return []
        try:
            methods: Tuple[XIdlMethod, ...] = fname.getMethods()
            lst = []
            for meth in methods:
                lst.append(meth.getName())
            lst.sort()
            return lst
        except Exception as e:
            raise Exception(f"Could not get Methods for: {interface_name}") from e

    @classmethod
    def show_methods(cls, interfce_name: str) -> None:
        """
        Prints methods to console for an interface

        Args:
            interfce_name (str): name of interface
        """
        methods = cls.get_methods(interface_name=interfce_name)
        if len(methods) == 0:
            print(f"No methods found for '{interfce_name}'")
            return
        print(f"{interfce_name} Methods: {len(methods)}")
        for method in methods:
            print(f"  {method}")

    @classmethod
    def show_methods_obj(cls, obj: object, property_concept: PropertyConceptEnum | None = None) -> None:
        """
        Prints method to console for an object such as a doc.

        Args:
            obj (object): Object to get methods of.
            property_concept (PropertyConceptEnum | None, optional): Type of method to get. Defaults to PropertyConceptEnum.ALL.
        """
        methods = cls.get_methods_obj(obj=obj, property_concept=property_concept)
        if len(methods) == 0:
            print(f"No Object methods found")
            return
        print(f"Object Methods: {len(methods)}")
        for method in methods:
            print(f"  {method}")

    # -------------------------- style info --------------------------
    @staticmethod
    def get_style_families(doc: object) -> XNameAccess:
        """
        Gets a list of style family names

        Args:
            doc (object): office document

        Raises:
            MissingInterfaceError: If Doc does not implement XStyleFamiliesSupplier interface
            Exception: If unable to get style Families

        Returns:
            XNameAccess: Style Families
        """
        try:
            xsupplier = mLo.Lo.qi(XStyleFamiliesSupplier, doc)
            if xsupplier is None:
                raise mEx.MissingInterfaceError(XStyleFamiliesSupplier)
            return xsupplier.getStyleFamilies()
        except mEx.MissingInterfaceError:
            raise
        except Exception as e:
            raise Exception("Unable to get family style names") from e

    @classmethod
    def get_style_family_names(cls, doc: object) -> List[str]:
        """
        Gets a list of style family names

        Args:
            doc (object): office document

        Raises:
            Exception: If unable to names.

        Returns:
            List[str]: List of style names
        """
        try:
            name_acc = cls.get_style_families(doc)
            names = name_acc.getElementNames()
            lst = list(names)
            lst.sort()
            return lst
        except Exception as e:
            raise Exception("Unable to get family style names") from e

    @classmethod
    def get_style_container(cls, doc: object, family_style_name: str) -> XNameContainer:
        """
        Gets style container of document for a family of styles

        Args:
            doc (object): office document
            family_style_name (str): Family style name

        Raises:
            MissingInterfaceError: if unable to obtain XNameContainer interface

        Returns:
            XNameContainer: Style Family container
        """
        name_acc = cls.get_style_families(doc)
        xcontianer = mLo.Lo.qi(XNameContainer, name_acc.getByName(family_style_name), raise_err=True)
        return xcontianer

    @classmethod
    def get_style_names(cls, doc: object, family_style_name: str) -> List[str]:
        """
        Gets a list of style names

        Args:
            doc (object): office document
            family_style_name (str): name of family style

        Raises:
            Exception: If unable to access Style names

        Returns:
            List[str]: List of style names
        """
        try:
            style_container = cls.get_style_container(doc=doc, family_style_name=family_style_name)
            names = style_container.getElementNames()
            lst = list(names)
            lst.sort()
            return lst
        except Exception as e:
            raise Exception("Could not access style names") from e

    @classmethod
    def get_style_props(cls, doc: object, family_style_name: str, prop_set_nm: str) -> XPropertySet:
        """
        Get style properties for a family of styles

        Args:
            doc (object): office document
            family_style_name (str): name of family style
            prop_set_nm (str): property set name

        Raises:
            MissingInterfaceError: if a required interface cannot be obtained.

        Returns:
            XPropertySet: Property set
        """
        style_container = cls.get_style_container(doc, family_style_name)
        #       container is a collection of named property sets
        name_props = mLo.Lo.qi(XPropertySet, style_container.getByName(prop_set_nm))
        if name_props is None:
            raise mEx.MissingInterfaceError(XPropertySet)
        return name_props

    @classmethod
    def get_page_style_props(cls, doc: object) -> XPropertySet:
        """
        Gets style properties for page styles

        Args:
            doc (object): office docs

        Raises:
            MissingInterfaceError: if a required interface cannot be obtained.

        Returns:
            XPropertySet: property set
        """
        return cls.get_style_props(doc, "PageStyles", "Standard")

    @classmethod
    def get_paragraph_style_props(cls, doc: object) -> XPropertySet:
        """
        Gets style properties for paragraph styles

        Args:
            doc (object): office docs

        Raises:
            MissingInterfaceError: if a required interface cannot be obtained.

        Returns:
            XPropertySet: property set
        """
        return cls.get_style_props(doc, "ParagraphStyles", "Standard")

    # ----------------------------- document properties ----------------------

    @classmethod
    def print_doc_properties(cls, doc: object) -> None:
        """
        Prints document properties to console

        Args:
            doc (object): office document
        """
        try:
            doc_props_supp = mLo.Lo.qi(XDocumentPropertiesSupplier, doc, True)
            dps = doc_props_supp.getDocumentProperties()
            cls.print_doc_props(dps=dps)
            ud_props = dps.getUserDefinedProperties()
            mProps.Props.show_obj_props("UserDefined Info", ud_props)
        except Exception as e:
            mLo.Lo.print("Unable to get doc properties")
            mLo.Lo.print(f"    {e}")
        return

    @classmethod
    def print_doc_props(cls, dps: XDocumentProperties) -> None:
        """
        Prints doc properties to console

        Args:
            dps (XDocumentProperties): document properties.

        See Also:
            :py:meth:`~Info.print_doc_properties`
        """
        print("Document Properties Info")
        print("  Author: " + dps.Author)
        print("  Title: " + dps.Title)
        print("  Subject: " + dps.Subject)
        print("  Description: " + dps.Description)
        print("  Generator: " + dps.Generator)

        keys = dps.Keywords
        print("  Keywords: ")
        for keyword in keys:
            print(f"  {keyword}")

        print("  Modified by: " + dps.ModifiedBy)
        print("  Printed by: " + dps.PrintedBy)
        print("  Template Name: " + dps.TemplateName)
        print("  Template URL: " + dps.TemplateURL)
        print("  Autoload URL: " + dps.AutoloadURL)
        print("  Default Target: " + dps.DefaultTarget)

        l = dps.Language
        loc = []
        loc.append("unknown" if len(l.Language) == 0 else l.Language)
        loc.append("unknown" if len(l.Country) == 0 else l.Country)
        loc.append("unknown" if len(l.Variant) == 0 else l.Variant)
        print(f"  Locale: {'; '.join(loc)}")

        print("  Modification Date: " + mDate.DateUtil.str_date_time(dps.ModificationDate))
        print("  Creation Date: " + mDate.DateUtil.str_date_time(dps.CreationDate))
        print("  Print Date: " + mDate.DateUtil.str_date_time(dps.PrintDate))
        print("  Template Date: " + mDate.DateUtil.str_date_time(dps.TemplateDate))

        doc_stats = dps.DocumentStatistics
        print("  Document statistics:")
        for nv in doc_stats:
            print(f"  {nv.Name} = {nv.Value}")

        try:
            print(f"  Autoload Secs: {dps.AutoloadSecs}")
        except Exception as e:
            print(f"  Autoload Secs: {e}")
        try:
            print(f"  Editing Cycles: {dps.EditingCycles}")
        except Exception as e:
            print(f"  Editing Cycles: {e}")
        try:
            print(f"  Editing Duration: {dps.EditingDuration}")
        except Exception as e:
            print(f"  Editing Duration: {e}")
        print()

    @staticmethod
    def set_doc_props(doc: object, subject: str, title: str, author: str) -> None:
        """
        Set document properties for subject, title, author

        Args:
            doc (object): office document
            subject (str): subject
            title (str): title
            author (str): author

        Raises:
            PropertiesError: If unable to set properties
        """
        try:
            dp_supplier = mLo.Lo.qi(XDocumentPropertiesSupplier, doc, True)
            doc_props = dp_supplier.getDocumentProperties()
            doc_props.Subject = subject
            doc_props.Title = title
            doc_props.Author = author
        except Exception as e:
            raise mEx.PropertiesError("Unable to set doc properties") from e

    @staticmethod
    def get_user_defined_props(doc: object) -> XPropertyContainer:
        """
        Gets user defined properties

        Args:
            doc (object): office document

        Raises:
            PropertiesError: if unable to access properties

        Returns:
            XPropertyContainer: Property container
        """
        try:
            dp_supplier = mLo.Lo.qi(XDocumentPropertiesSupplier, doc)
            if dp_supplier is None:
                raise mEx.MissingInterfaceError(XDocumentPropertiesSupplier)
            dps = dp_supplier.getDocumentProperties()
            return dps.getUserDefinedProperties()
        except Exception as e:
            raise mEx.PropertiesError("Unable to get user defined props") from e

    # ----------- installed package info -----------------

    @staticmethod
    def get_pip() -> XPackageInformationProvider:
        """
        Gets Package Information Provider

        Raises:
            MissingInterfaceError: if unable to obtain XPackageInformationProvider interface

        Returns:
            XPackageInformationProvider: Package Information Provider
        """
        ctx = mLo.Lo.get_context()
        pip = mLo.Lo.qi(
            XPackageInformationProvider,
            ctx.getValueByName("/singletons/com.sun.star.deployment.PackageInformationProvider"),
            True,
        )
        return pip
        # return pip.get(mLo.Lo.get_context())

    @classmethod
    def list_extensions(cls) -> None:
        """
        Prints extensions to console
        """
        try:
            pip = cls.get_pip()
        except mEx.MissingInterfaceError:
            print("No package info provider found")
            return
        exts_tbl = pip.getExtensionList()
        print("\nExtensions:")
        for i in range(len(exts_tbl)):
            print(f"{i+1}. ID: {exts_tbl[i][0]}")
            print(f"   Version: {exts_tbl[i][1]}")
            print(f"   Loc: {pip.getPackageLocation(exts_tbl[i][0])}")
            print()

    @classmethod
    def get_extension_info(cls, id: str) -> Tuple[str, ...]:
        """
        Gets infor for an installed extension in LibreOffice.

        Args:
            id (str): Extension id

        Returns:
            Tuple[str, ...]: Extension info
        """
        try:
            pip = cls.get_pip()
        except mEx.MissingInterfaceError:
            mLo.Lo.print("No package info provider found")
            return ()
        exts_tbl = pip.getExtensionList()
        mLo.Lo.print_table("Extension", exts_tbl)
        for el in exts_tbl:
            if el[0] == id:
                return el

        mLo.Lo.print(f"Extension {id} is not found")
        return ()

    @classmethod
    def get_extension_loc(cls, id: str) -> str | None:
        """
        Gets location for an installed extension in LibreOffice

        Args:
            id (str): Extension id

        Returns:
            str | None: Extension location on success; Otherwise, None
        """
        try:
            pip = cls.get_pip()
        except mEx.MissingInterfaceError:
            mLo.Lo.print("No package info provider found")
            return None
        return pip.getPackageLocation(id)

    @staticmethod
    def get_filter_names() -> Tuple[str, ...]:
        """
        Gets filter names

        Returns:
            Tuple[str, ...]: Filter names
        """
        na = mLo.Lo.create_instance_mcf(XNameAccess, "com.sun.star.document.FilterFactory")
        if na is None:
            mLo.Lo.print("No Filter factory found")
            return ()
        return na.getElementNames()

    @staticmethod
    def get_filter_props(filter_nm: str) -> List[PropertyValue]:
        """
        Gets filter properties

        Args:
            filter_nm (str): Filter Name

        Returns:
            List[PropertyValue]: List of PropertValue
        """
        na = mLo.Lo.create_instance_mcf(XNameAccess, "com.sun.star.document.FilterFactory")
        if na is None:
            mLo.Lo.print("No Filter factory found")
            return []
        result = na.getByName(filter_nm)
        if result is None:
            mLo.Lo.print(f"No props for filter: {filter_nm}")
            return []
        return list(result)

    @classmethod
    def is_import(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.IMPORT`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.IMPORT) == cls.Filter.IMPORT

    @classmethod
    def is_export(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.EXPORT`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.EXPORT) == cls.Filter.EXPORT

    @classmethod
    def is_template(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.TEMPLATE`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.TEMPLATE) == cls.Filter.TEMPLATE

    @classmethod
    def is_internal(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.INTERNAL`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.INTERNAL) == cls.Filter.INTERNAL

    @classmethod
    def is_template_path(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.TEMPLATEPATH`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.TEMPLATEPATH) == cls.Filter.TEMPLATEPATH

    @classmethod
    def is_own(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.OWN`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.OWN) == cls.Filter.OWN

    @classmethod
    def is_alien(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.ALIEN`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.ALIEN) == cls.Filter.ALIEN

    @classmethod
    def is_default(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.DEFAULT`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.DEFAULT) == cls.Filter.DEFAULT

    @classmethod
    def is_support_selection(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.SUPPORTSSELECTION`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.SUPPORTSSELECTION) == cls.Filter.SUPPORTSSELECTION

    @classmethod
    def is_not_in_file_dialog(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.NOTINFILEDIALOG`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.NOTINFILEDIALOG) == cls.Filter.NOTINFILEDIALOG

    @classmethod
    def is_not_in_chooser(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.NOTINCHOOSER`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.NOTINCHOOSER) == cls.Filter.NOTINCHOOSER

    @classmethod
    def is_read_only(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.READONLY`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.READONLY) == cls.Filter.READONLY

    @classmethod
    def is_third_party_filter(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.THIRDPARTYFILTER`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.THIRDPARTYFILTER) == cls.Filter.THIRDPARTYFILTER

    @classmethod
    def is_preferred(cls, filter_flags: Info.Filter) -> bool:
        """
        Gets if filter flags has ``Filter.PREFERRED`` flag set

        Args:
            filter_flags (Filter): Flags

        Returns:
            bool: True if flag is set; Otherwise, False
        """
        return (filter_flags & cls.Filter.PREFERRED) == cls.Filter.PREFERRED

    @staticmethod
    def is_type_struct(obj: object, type_name: str) -> bool:
        """
        Gets if an object is a Uno Struct of matching type.

        Args:
            obj (object): Object to test if is struct
            type_name (str): Type string such as 'com.sun.star.table.CellRangeAddress'

        Returns:
            bool: True if 'obj' is struct and 'obj' matches 'type_name'; Otherwise, False
        """
        if obj is None:
            return False
        if hasattr(obj, "typeName"):
            return obj.typeName == type_name
        return False

    @staticmethod
    def is_type_interface(obj: object, type_name: str) -> bool:
        """
        Gets if an object is a Uno interface of matching type.

        Args:
            obj (object): Object to test if is interface
            type_name (str): Type string such as 'com.sun.star.uno.XInterface'

        Returns:
            bool: True if 'obj' is interface and 'obj' matches 'type_name'; Otherwise, False
        """
        return mLo.Lo.is_uno_interfaces(obj, type_name)
        # if obj is None:
        #     return False
        # if hasattr(obj, "__pyunointerface__"):
        #     return obj.__pyunointerface__ == type_name
        # elif hasattr(obj, "queryInterface"):
        #     uno_t = uno.getTypeByName(type_name)
        #     q_obj = obj.queryInterface(uno_t)
        #     if q_obj is not None:
        #         return True
        # return False

    @staticmethod
    def is_type_enum(obj: uno.Enum, type_name: str) -> bool:
        """
        Gets if an object is a Uno enum of matching type.

        Args:
            obj (object): Object to test if is uno enum
            type_name (str): Type string such as ``com.sun.star.sheet.GeneralFunction``

        Returns:
            bool: True if ``obj`` is uno enum and ``obj`` matches ``type_name``; Otherwise, False
        """
        if obj is None:
            return False
        if hasattr(obj, "typeName"):
            return obj.typeName == type_name
        return False

    # region is_type_enum_multi()
    @overload
    @staticmethod
    def is_type_enum_multi(alt_type: str, enum_type: Type[Enum], enum_val: Enum) -> bool:
        ...

    @overload
    @staticmethod
    def is_type_enum_multi(alt_type: str, enum_type: Type[Enum], enum_val: Enum, arg_name: str) -> bool:
        ...

    @staticmethod
    def is_type_enum_multi(alt_type: str, enum_type: Type[Enum], enum_val: Enum, arg_name: str = "") -> bool:
        """
        Gets if an multiple inheritance enum, such as a ``str, Enum`` is of expected type.

        Args:
            alt_type (str): Alternative Type, In the case of a ``str, Enum`` this would be ``str``
            enum_type (Type[Enum]): Expected Enum Type
            enum_val (Enum): Actual Enum Value
            arg_name (str, optional): Argument name used to pass enum value into method.

        Raises:
            TypeError: If ``arg_name`` is passed into this method and the type check fails.

        Returns:
            bool: ``True`` if ``enum_val`` is valid ``alt_type`` or ``enum_type``; Otherwise, ``False``

        Example:
            .. code-block:: python

                >>> from ooodev.utils.kind import chart2_types as ct

                >>> val = ct.Area3dKind.PERCENT_STACKED_AREA_3D
                >>> print(is_enum_type("str", ct.ChartTemplateBase, val))
                True
                >>> print(is_enum_type("str", ct.ChartTypeNameBase, val))
                False
                >>> print(is_enum_type("str", ct.ChartTypeNameBase, val, "input_enum"))
                TypeError: Parameter "input_enum" must be of type "str" or "ChartTypeNameBase"
        """
        if type(enum_val).__name__ != alt_type:
            if not isinstance(enum_val, enum_type):
                if arg_name:
                    name = enum_type.__name__
                    raise TypeError(f'Parameter "{arg_name}" must be of type "{alt_type}" or "{name}"')
                else:
                    return False
        return True

    # endregion is_type_enum_multi()

    @classmethod
    def get_type_name(cls, obj: object) -> str | None:
        """
        Gets type name such as ``com.sun.star.table.TableSortField`` from uno object.

        Args:
            obj (object): Uno object

        Returns:
            str | None: Full type name if found; Otherwise; None
        """
        if hasattr(obj, "typeName"):
            return obj.typeName
        if hasattr(obj, "__ooo_full_ns__"):
            # ooouno object
            return obj.__ooo_full_ns__
        if hasattr(obj, "__pyunointerface__"):
            return obj.__pyunointerface__
        return None

    @classproperty
    def language(cls) -> str:
        """
        Gets the Current Language of the LibreOffice Instance

        Returns:
            str: First two chars of language in lower case such as 'en-US'
        """

        try:
            return cls._language
        except AttributeError:
            lang = cls.get_config(node_str="ooLocale")
            cls._language = str(lang)
        return cls._language

    @language.setter
    def language(cls, value) -> None:
        # raise error on set. Not really necessary but gives feedback.
        raise AttributeError("Attempt to modify read-only class property '%s'." % cls.__name__)

    @classproperty
    def version(cls) -> str:
        """
        Gets the running LibreOffice version

        Returns:
            str: version as string such as ``"7.3.4.2"``

        Note:
            This property only works after Office is Loaded.
        """

        try:
            return cls._version
        except AttributeError:
            # ooSetupVersionAboutBox long "7.3.4.2"
            # ooSetupVersion short "7.3"
            lang = cls.get_config(node_str="ooSetupVersionAboutBox")
            cls._version = str(lang)
        return cls._version

    @version.setter
    def version(cls, value) -> None:
        # raise error on set. Not really necessary but gives feedback.
        raise AttributeError("Attempt to modify read-only class property '%s'." % cls.__name__)

    @classproperty
    def version_info(cls) -> Tuple[int, ...]:
        """
        Gets the running LibreOffice version

        Returns:
            tuple: version as tuple such as ``(7, 3, 4, 2)``

        Note:
            This property only works after Office is Loaded.
        """

        try:
            return cls._version_info
        except AttributeError:
            cls._version_info = tuple([int(s) for s in cls.version.split(".")])
        return cls._version_info

    @version_info.setter
    def version_info(cls, value) -> None:
        # raise error on set. Not really necessary but gives feedback.
        raise AttributeError("Attempt to modify read-only class property '%s'." % cls.__name__)


def _del_cache_attrs(source: object, e: EventArgs) -> None:
    # clears Write Attributes that are dynamically created
    dattrs = ("_language", "_version", "_version_info")
    for attr in dattrs:
        if hasattr(Info, attr):
            delattr(Info, attr)


# subscribe to events that warrant clearing cached attribs
_Events().on(LoNamedEvent.RESET, _del_cache_attrs)

__all__ = ("Info",)



class GlobalEditDispatch:
    """
    Global Edit Dispatch Commands

    See Also:
        - :py:meth:`.Lo.dispatch_cmd`
        - `Global Dispatch commands <https://wiki.documentfoundation.org/Development/DispatchCommands#Global>`_
    """

    COMPARE_DOCUMENTS = "CompareDocuments"
    """
    Compare Non-Track Changed Document
    
    Args
        - ``URL (str)``
        - ``FilterName (str)``
        - ``Password (str)``
        - ``FilterOptions (str)``
        - ``Version (int)``
        - ``NoAcceptDialog (bool)``
    """
    COPY = "Copy"
    COPY_HYPERLINK_LOCATION = "CopyHyperlinkLocation"
    """Copy Hyperlink Location"""
    CUT = "Cut"
    DS_BROWSER_EXPLORER = "DSBrowserExplorer"
    """Explorer On/Off"""
    DELETE = "Delete"
    """Delete Contents"""
    DELETE_ALL_NOTES = "DeleteAllNotes"
    """Delete All Comments"""
    DELETE_AUTHOR = "DeleteAuthor"
    """Delete All Comments by This Author"""
    DELETE_COMMENT = "DeleteComment"
    """
    Delete Comment

    Args
        - ``ID (postitid)``
    """
    DELETE_COMMENT_THREAD = "DeleteCommentThread"
    """
    Delete Comment Thread

    Args
        - ``ID (postitid)``
    """
    DELETE_FRAME = "DeleteFrame"
    """Delete Frame"""
    EDIT_HYPERLINK = "EditHyperlink"
    """Edit Hyperlink"""
    EDIT_STYLE = "EditStyle"
    """
    Edit

    Args
        - ``Param (str)``
        - ``Family (int)``
    """
    FORMAT_ALL_NOTES = "FormatAllNotes"
    """Format All Comments"""
    GET_COLOR_TABLE = "GetColorTable"
    """Color Palette"""
    HYPERLINK_DIALOG = "HyperlinkDialog"
    """Insert Hyperlink"""
    INSERT_MODE = "InsertMode"
    """Insert Mode"""
    LEAVE_FM_CREATE_MODE = "LeaveFMCreateMode"
    MERGE_DOCUMENTS = "MergeDocuments"
    """
    Merge Track Changed Document

    Args
        - ``URL (str)``
        - ``Version (int)``
    """
    OBJECT_MENUE = "ObjectMenue"
    """Menu for editing or saving OLE objects"""
    OPEN_HYPERLINK_ON_CURSOR = "OpenHyperlinkOnCursor"
    """Open Hyperlink"""
    OPEN_SMART_TAG_MENU_ON_CURSOR = "OpenSmartTagMenuOnCursor"
    """Smart Tags"""
    PASTE = "Paste"
    PASTE_UNFORMATTED = "PasteUnformatted"
    """``Un-formatted`` Text"""
    PROTECT_TRACE_CHANGE_MODE = "ProtectTraceChangeMode"
    """Protect Track Changes"""
    REDO = "Redo"
    REFRESH = "Refresh"
    REPEAT = "Repeat"
    REPLY_COMMENT = "ReplyComment"
    """Reply Comment"""
    RESOLVE_COMMENT = "ResolveComment"
    """Resolved"""
    RESOLVE_COMMENT_THREAD = "ResolveCommentThread"
    """Resolved Thread"""
    SEARCH_DIALOG = "SearchDialog"
    """Find and Replace"""
    SEARCH_FORMATTED_DISPLAY_STRING = "SearchFormattedDisplayString"
    """Search Formatted Display String"""
    SEARCH_LABEL = "SearchLabel"
    """[placeholder for message]"""
    SELECT = "Select"
    """Select All"""
    SELECT_ALL = "SelectAll"
    """Select All"""
    SELECT_OBJECT = "SelectObject"
    """Select"""
    SPELL_DIALOG = "SpellDialog"
    """Spelling"""
    SPELLING_AND_GRAMMAR_DIALOG = "SpellingAndGrammarDialog"
    """Check Spelling"""
    SPELLING_DIALOG = "SpellingDialog"
    """Check Spelling"""
    UNDO = "Undo"
    UNDO_ACTION = "UndoAction"



from abc import abstractmethod
from enum import Enum


class KindBase(str, Enum):
    def __str__(self) -> str:
        return self.value
    
    @abstractmethod
    def to_namespace(self) -> str:
        ...


from dataclasses import dataclass


@dataclass(frozen=True)
class WindowTitle:
    """Window Title Info"""

    title: str
    """Window Title"""
    is_regex: bool = False
    """Determines if title is treated as regular expression."""
    class_name: str = "SALFRAME"



# coding: utf-8
"""General Utilities"""
from __future__ import annotations
import re
from typing import Iterable, Iterator, NamedTuple, Any
from inspect import isclass

# match:
#   Any uppercase character that is not at the start of a line
#   Any Number that is preceeded by a Upper or Lower case character
_REG_TO_SNAKE = re.compile(r"(?<!^)(?=[A-Z])|(?<=[A-zA-Z])(?=[0-9])")  # re.compile(r"(?<!^)(?=[A-Z])")
_REG_LETTER_AFTER_NUMBER = re.compile(r"(?<=\d)(?=[a-zA-Z])")


class ArgsHelper:
    "Args Helper"

    class NameValue(NamedTuple):
        "Name Value pair"
        name: str
        """Name component"""
        value: Any
        """Value component"""


class Util:
    @classmethod
    def is_iterable(cls, arg: object, excluded_types: Iterable[type] | None = None) -> bool:
        """
        Gets if ``arg`` is iterable.

        Args:
            arg (object): object to test
            excluded_types (Iterable[type], optional): Iterable of type to exclude.
                If ``arg`` matches any type in ``excluded_types`` then ``False`` will be returned.
                Default ``(str,)``

        Returns:
            bool: ``True`` if ``arg`` is an iterable object and not of a type in ``excluded_types``;
            Otherwise, ``False``.

        Note:
            if ``arg`` is of type str then return result is ``False``.

        .. collapse:: Example

            .. code-block:: python

                # non-string iterables
                assert is_iterable(arg=("f", "f"))       # tuple
                assert is_iterable(arg=["f", "f"])       # list
                assert is_iterable(arg=iter("ff"))       # iterator
                assert is_iterable(arg=range(44))        # generator
                assert is_iterable(arg=b"ff")            # bytes (Python 2 calls this a string)

                # strings or non-iterables
                assert not is_iterable(arg=u"ff")        # string
                assert not is_iterable(arg=44)           # integer
                assert not is_iterable(arg=is_iterable)  # function

                # excluded_types, optionally exlcude types
                from enum import Enum, auto

                class Color(Enum):
                    RED = auto()
                    GREEN = auto()
                    BLUE = auto()

                assert is_iterable(arg=Color)             # Enum
                assert not is_iterable(arg=Color, excluded_types=(Enum, str)) # Enum
        """
        # if isinstance(arg, str):
        #     return False
        if excluded_types is None:
            excluded_types = (str,)
        result = False
        try:
            result = isinstance(iter(arg), Iterator)
        except Exception:
            result = False
        if result is True:
            if cls._is_iterable_excluded(arg, excluded_types=excluded_types):
                result = False
        return result

    @staticmethod
    def _is_iterable_excluded(arg: object, excluded_types: Iterable) -> bool:
        try:
            isinstance(iter(excluded_types), Iterator)
        except Exception:
            return False

        if len(excluded_types) == 0:
            return False

        def _is_instance(obj: object) -> bool:
            # when obj is instance then isinstance(obj, obj) raises TypeError
            # when obj is not instance then isinstance(obj, obj) return False
            try:
                if not isinstance(obj, obj):
                    return False
            except TypeError:
                pass
            return True

        ex_types = excluded_types if isinstance(excluded_types, tuple) else tuple(excluded_types)
        arg_instance = _is_instance(arg)
        if arg_instance is True:
            if isinstance(arg, ex_types):
                return True
            return False
        if isclass(arg) and issubclass(arg, ex_types):
            return True
        return arg in ex_types

    @staticmethod
    def to_camel_case(s: str) -> str:
        """
        Converts string to ``CamelCase``

        Args:
            s (str): string to convert such as ``snake_case_word`` or ``pascalCaseWord``

        Returns:
            str: string converted to ``CamelCaseWord``
        """
        s = s.strip()
        if not s:
            return ""
        result = s
        if "_" in s:
            result = "".join(word.title() for word in s.split("_"))
            return result
        # convert to CamelCase if pascalCase was passed in.
        result = result[:1].upper() + result[1:]
        return result

    @classmethod
    def to_pascal_case(cls, s: str) -> str:
        """
        Converts string to ``pascalCase``

        Args:
            s (str): string to convert such as ``snake_case_word`` or  ``CamelCaseWord``

        Returns:
            str: string converted to ``pascalCaseWord``
        """
        result = cls.to_camel_case(s)
        if result:
            result = result[:1].lower() + result[1:]
        return result

    @staticmethod
    def to_snake_case(s: str) -> str:
        """
        Convert string to ``snake_case``

        Args:
            s (str): string to convert such as ``pascalCaseWord`` or  ``CamelCaseWord``

        Returns:
            str: string converted to ``snake_case_word``
        """
        s = s.strip()
        if not s:
            return ""
        result = _REG_TO_SNAKE.sub("_", s)
        result = _REG_LETTER_AFTER_NUMBER.sub("_", result)
        return result.lower()

    @classmethod
    def to_snake_case_upper(cls, s: str) -> str:
        """
        Convert string to ``SNAKE_CASE``

        Args:
            s (str): string to convert such as ``snake_case_word`` or ``pascalCaseWord`` or  ``CamelCaseWord``

        Returns:
            str: string converted to ``SNAKE_CASE_WORD``
        """
        result = cls.to_snake_case(s)
        if not s:
            return ""
        return result.upper()

    @staticmethod
    def to_single_space(s: str, strip=True) -> str:
        """
        Gets a string with multiple spaces converted to single spaces

        Args:
            s (str): String
            strip (bool, optional): If ``True`` then whitespace is stripped from start and end or string. Default ``True``.

        Returns:
            str: String with extra spaces removed.

        Example:
            .. code-block:: python

                >>> s = ' The     quick brown    fox'
                >>> print(Util.to_single_space(s))
                'The quick brown fox'
        """
        if not s:
            return ""
        result = re.sub(" +", " ", s)
        if strip:
            return result.strip()
        return result



# coding: utf-8
"""make/get/set properties in an array"""
# Python conversion of Props.java by Andrew Davison, ad@fivedots.coe.psu.ac.th
# See Also: https://fivedots.coe.psu.ac.th/~ad/jlop/
# region Imports
from __future__ import annotations
from typing import Any, Iterable, List, Optional, Sequence, Tuple, TYPE_CHECKING, cast, overload
import uno

from com.sun.star.beans import PropertyAttribute  # const
from com.sun.star.beans import XPropertySet
from com.sun.star.beans import XFastPropertySet
from com.sun.star.container import XNameAccess
from com.sun.star.document import XTypeDetection
from com.sun.star.ui import ItemType  # const
from com.sun.star.ui import ItemStyle  # const
from com.sun.star.beans import PropertyVetoException
from com.sun.star.beans import UnknownPropertyException
from com.sun.star.container import XIndexAccess


if TYPE_CHECKING:
    from com.sun.star.beans import Property, PropertyValue, XPropertySetInfo
    from com.sun.star.beans import XMultiPropertySet

    # import module and not module content to avoid circular import issue.
    # https://stackoverflow.com/questions/22187279/python-circular-importing
    #
    # using lazy loading: https://snarky.ca/lazy-importing-in-python-3-7/
from . import lo as mLo
from . import info as mInfo

from ..exceptions import ex as mEx

# endregion Imports


class Props:
    """make/get/set properties in an array"""

    # region ------------------- make properties -----------------------
    @staticmethod
    def make_prop_value(name: Optional[str] = None, value: Optional[str] = None) -> PropertyValue:
        """
        Makes a Uno Property Value and assigns name and value if present.

        Args:
            name (Optional[str], optional): Property name
            value (Optional[str], optional): Property value

        Returns:
            PropertyValue: com.sun.star.beans.PropertyValue

        See Also:
            `LibreOffice API PropertyValue <https://api.libreoffice.org/docs/idl/ref/structcom_1_1sun_1_1star_1_1beans_1_1PropertyValue.html>`_
        """
        p: PropertyValue = uno.createUnoStruct("com.sun.star.beans.PropertyValue")
        if name is not None:
            p.Name = name
        if value is not None:
            p.Value = value
        return p

    @classmethod
    def make_bar_item(
        cls, cmd: str, item_name: str
    ) -> Tuple[PropertyValue, PropertyValue, PropertyValue, PropertyValue, PropertyValue]:
        """
        Properties for a toolbar item using a name and an image

        problem: image does not appear next to text on toolbar

        Args:
            cmd (str): Value of CommandURL
            item_name (str): Label assigned to bar

        Returns:
            Tuple[PropertyValue, PropertyValue, PropertyValue, PropertyValue, PropertyValue]: Tuple of properties. (CommandURL, Label, Type, Visible, Style)
        """
        p1 = cls.make_prop_value(name="CommandURL", value=cmd)
        p2 = cls.make_prop_value(name="Label", value=item_name)
        p3 = cls.make_prop_value(name="Type", value=ItemType.DEFAULT)
        p4 = cls.make_prop_value(name="Visible", value=True)
        p5 = cls.make_prop_value(name="Style")
        p5.Value = ItemStyle.DRAW_FLAT + ItemStyle.ALIGN_LEFT + ItemStyle.AUTO_SIZE + ItemStyle.ICON + ItemStyle.TEXT

        return (p1, p2, p3, p4, p5)

    @classmethod
    def make_props(cls, **kwargs) -> Tuple[PropertyValue, ...]:
        """
        Make Properties

        Keyword Args:
            kwargs (Dict[str, Any]): Each key, value pair is assigned to a PropertyValue.

        Returns:
            Tuple[PropertyValue, ...]: Tuple of Properties

        Note:
            String properties such as ``Zoom.Value`` can be pass by constructing a dictionary
            and passing dictionary via expansion.

            Example Expansion

            .. code-block:: python

                p_dic = {
                    "Zoom.Value": 0,
                    "Zoom.ValueSet": 28703,
                    "Zoom.Type": 1
                    }
                props = Props.make_props(**p_dic)
        """
        lst = []
        for k, v in kwargs.items():
            lst.append(cls.make_prop_value(name=k, value=v))
        return tuple(lst)

    # endregion ---------------- make properties -----------------------

    # region ------------------- uno -----------------------------------
    @staticmethod
    def any(*elements: object) -> uno.Any:
        """
        Gets a uno.Any object for elements.

        The first element determines the type for the uno.Any object.

        Raises:
            ValueError: if unable to create uno.Any object.

        Returns:
            uno.Any: uno.Any Object

        Note:
            uno.Any is usually constructed in the following manor.

            ``uno.Any("[]com.sun.star.table.TableSortField", (sort_one, sort_two)``

            This method is a shortcut.

            ``Props.any(sort_one, sort_two)``
        """
        if len(elements) == 0:
            raise ValueError("No args to create unn.Any object")
        obj = elements[0]
        if isinstance(obj, uno.Type):
            type_name = obj.typeName
            return uno.Any(obj, [*elements])
        elif isinstance(obj, str):
            type_name = "string"
        elif isinstance(obj, int):
            type_name = "short"
        else:
            type_name = mInfo.Info.get_type_name(obj)
        if type_name is None:
            raise ValueError("Unable to get type name to create uno.Any object")
        return uno.Any(f"[]{type_name}", [*elements])

    # endregion ---------------- uno -----------------------------------

    # region ------------------- set properties ------------------------
    @staticmethod
    def set_prop(props: Iterable[PropertyValue], name: str, value: object) -> bool:
        """
        Sets property value for the first property that has matching name.

        Args:
            props (Iterable[PropertyValue]): Property Values
            name (str): Property name
            value (object): Property Value

        Raises:
            TypeError: if props is None.

        Returns:
            bool: True if property matching name has been updated; Otherwise, False
        """
        if props is None:
            TypeError(f"Property array is null; cannot set {name}")
        for prop in props:
            if prop.Name == name:
                prop.Value = value
                return True
        print(f"{name} not found")
        return False

    @staticmethod
    def get_xproperty_set(obj: object) -> XPropertySet:
        """
        Gets Property Set

        Args:
            obj (object): object that implements ``XPropertySet``

        Raises:
            MissingInterfaceError: if ``obj`` does not implement ``XPropertySet``.

        Returns:
            XPropertySet: Property Set
        """
        return mLo.Lo.qi(XPropertySet, obj, True)

    @staticmethod
    def get_xproperty_set_fast(obj: object) -> XFastPropertySet:
        """
        Gets Fast Property Set

        Args:
            obj (object): object that implements ``XFastPropertySet``

        Raises:
            MissingInterfaceError: if ``obj`` does not implement ``XFastPropertySet``.

        Returns:
            XFastPropertySet: Property Set
        """
        return mLo.Lo.qi(XFastPropertySet, obj, True)

    # region get_xproperty_fast_value()
    @overload
    @classmethod
    def get_xproperty_fast_value(cls, *, handle: int, fps: XFastPropertySet) -> Any:
        ...

    @overload
    @classmethod
    def get_xproperty_fast_value(cls, *, prop: Property, fps: XFastPropertySet) -> Any:
        ...

    @overload
    @classmethod
    def get_xproperty_fast_value(cls, *, handle: int, obj: object) -> Any:
        ...

    @classmethod
    def get_xproperty_fast_value(cls, *args, **kwargs) -> Any:
        """
        Gets a fast property value via ``XFastPropertySet``

        Arguments:
            handle (int): handle of the property.
            fps (XFastPropertySet): property set that contains the property for matching handle.
            prop (Property): Property that handle can be obtained from.
            obj (object): an object that ``XFastPropertySet`` can be obtained from.

        Raises:
            PropertyGeneralError: If an error occurs.

        Returns:
            Any: Property Value
        """
        if len(args) > 0:
            raise TypeError(f"get_xproperty_fast_value() takes 0 positional arguments but {len(args)} was given")

        count = len(kwargs)

        if count != 2:
            raise TypeError("get_xproperty_fast_value() got an invalid number of arguments")

        try:
            handle = cast(int, kwargs.get("handle", None))
            if handle is None:
                handle = cast(Property, kwargs.get("prop")).Handle

            fps = cast(XFastPropertySet, kwargs.get("fps", None))
            if fps:
                return fps.getFastPropertyValue(handle)

            fps = cls.get_xproperty_set_fast(kwargs.get("obj"))
            return fps.getFastPropertyValue(handle)
        except Exception as e:
            raise mEx.PropertyGeneralError("Error getting fast property value") from e

    # endregion get_xproperty_fast_value()

    @classmethod
    def get_property_set_info(cls, obj: object) -> XPropertySetInfo:
        """
        Gets property set info

        Args:
            obj (object): Object that implements ``XPropertySet``.

        Raises:
            PropertyGeneralError: If error occurs

        Returns:
            XPropertySetInfo: Property Set Info
        """
        try:
            xset = cls.get_xproperty_set(obj)
            result = xset.getPropertySetInfo()
            if result is None:
                raise mEx.NoneError("None Value: xset.getPropertySetInfo() returned None")
            return result
        except Exception as e:
            raise mEx.PropertyGeneralError("Error getting property set info") from e

    @classmethod
    def get_property_by_name(cls, obj: object, name: str) -> Property:
        """
        Gets a property by name.

        Property instances do not contain a value.

        Args:
            obj (object): Object to get a property from
            name (str): Property name to lookup in ``obj``.

        Raises:
            PropertyError: If error occurs

        Returns:
            Property: Property instance.
        """
        try:
            info = cls.get_property_set_info(obj)
            result = info.getPropertyByName(name)
            if result is None:
                pass
            return result
        except Exception as e:
            raise mEx.PropertyError(name) from e

    @classmethod
    def has_property(cls, obj: object, name: str) -> bool:
        """
        Gets is an object contains a Property

        Args:
            obj (object): object that implements ``XPropertySet``
            name (str): _description_

        Returns:
            bool: ``True`` if object contains property with name; Otherwise, ``False``.
        """
        try:
            xset = cls.get_property_set_info(obj)
            return xset.hasPropertyByName(name)
        except Exception:
            pass
        return False

    @staticmethod
    def get_prop(props: Iterable[PropertyValue], name: str) -> object | None:
        """
        Gets property value for property that matches name.

        Args:
            props (Iterable[PropertyValue]): Properties to search
            name (str): Property name to find.

        Raises:
            TypeError: if props is None.

        Returns:
            object | None: Property value if found; Otherwise; None
        """
        if props is None:
            TypeError(f"Property array is null; cannot get {name}")
        for prop in props:
            if prop.Name == name:
                return prop.Value
        print(f"{name} not found")
        return None

    # region    set_property()
    @overload
    @classmethod
    def set_property(cls, obj: object, name: str, value: object) -> None:
        """
        Sets the value of the property with the specified name.

        Args:
            obj (object): object that implements XPropertySet interface
            name (str): Name of property to set value of
            value (object): property value

        Raises:
            MissingInterfaceError: If required interface cannot be obtained.
            Exception: If unable to set property value
        """
        ...

    @overload
    @classmethod
    def set_property(cls, prop_set: XPropertySet, name: str, value: object) -> None:
        """
        Sets the value of the property with the specified name.

        Args:
            prop_set (XPropertySet): Property set
            name (str): Name of property to set value of
            value (object): property value

        Raises:
            Exception: If unable to set property value
        """
        ...

    @classmethod
    def set_property(cls, *args, **kwargs) -> None:
        """
        Sets the value of the property with the specified name.

        Args:
            obj (object): object that implements XPropertySet interface
            prop_set (XPropertySet): Property set
            name (str): Name of property to set value of
            value (object): property value

        Raises:
            MissingInterfaceError: If required interface cannot be obtained.
            PropertySetError: If unable to set property value
        """
        ordered_keys = (1, 2, 3)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("obj", "prop_set", "name", "value")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("set_property() got an unexpected keyword argument")
            keys = ("obj", "prop_set")
            for key in keys:
                if key in kwargs:
                    ka[1] = kwargs[key]
                    break
            ka[2] = kwargs.get("name", None)
            ka[3] = kwargs.get("value", None)
            return ka

        if count != 3:
            raise TypeError("set_property() got an invalid numer of arguments")

        kargs = get_kwargs()
        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg

        try:

            cls.set(kargs[1], **{kargs[2]: kargs[3]})
        except mEx.MissingInterfaceError:
            raise
        except mEx.MultiError as me:
            err = me.errors[0]
            if isinstance(err, (mEx.PropertyNotFoundError, mEx.PropertySetError)):
                raise err
            raise mEx.PropertySetError(f"Could not set property '{kargs[2]}'") from err
        # if mLo.Lo.is_uno_interfaces(kargs[1], XPropertySet):
        #     prop_set = cast(XPropertySet, kargs[1])
        # else:
        #     prop_set = mLo.Lo.qi(XPropertySet, kargs[1], raise_err=True)
        # try:
        #     prop_set.setPropertyValue(kargs[2], kargs[3])
        # except Exception as e:
        #     raise mEx.PropertySetError(f"Could not set property '{kargs[2]}'") from e

    # endregion set_property()

    # region    set_properties()
    @overload
    @classmethod
    def set_properties(cls, obj: object, names: Sequence[str], vals: Sequence[object]) -> None:
        """
        Set Properties

        Args:
            obj (object): Object that implements XPropertySet interface
            names (Sequence[str]): Property Names
            vals (Sequence[object]): Property Values

        Raises:
            MissingInterfaceError: if obj does not implement XPropertySet interface
            MultiError: If unable to set a property
        """
        ...

    @overload
    @classmethod
    def set_properties(cls, prop_set: XPropertySet, names: Sequence[str], vals: Sequence[object]) -> None:
        """
        Set Properties

        Args:
            prop_set (XPropertySet): Property Set
            names (Sequence[str]): Property Names
            vals (Sequence[object]): Property Values

        Raises:
            MultiError: If unable to set a property
        """
        ...

    @overload
    @classmethod
    def set_properties(cls, obj: object, from_obj: object) -> None:
        """
        Set properties

        Properties of ``from_obj`` are assigned to ``obj``

        Args:
            obj (object): Object that implements XPropertySet interface
            from_obj (object): Other object that implements XPropertySet interface

        Raises:
            MissingInterfaceError: if obj does not implement XPropertySet interface
            MissingInterfaceError: if from_obj does not implement XPropertySet interface
            MultiError: If unable to set a property
        """
        ...

    @overload
    @classmethod
    def set_properties(cls, prop_set: XPropertySet, from_props: XPropertySet) -> None:
        """
        Set properties

        Properties of ``from_props`` are assigned to ``prop_set``

        Args:
            prop_set (XPropertySet): Property set
            from_props (XPropertySet): Property set

        Raises:
            MultiError: If unable to set a property
        """
        ...

    @classmethod
    def set_properties(cls, *args, **kwargs) -> None:
        """
        Set Properties

        Args:
            obj (object): Object that implements XPropertySet interface
            prop_set (XPropertySet): Property set
            from_props (XPropertySet): Property set
            from_obj (object): Other object that implements XPropertySet interface
            names (Sequence[str]): Property Names
            vals (Sequence[object]): Property Values

        Raises:
            MissingInterfaceError: if obj does not implement XPropertySet interface
            MissingInterfaceError: if from_obj does not implement XPropertySet interface
            MultiError: If unable to set a property

        Returns:
            None:

        Note:
            If ``MultiError`` occurs only the properties that raised an error is part of the error object.
            The remaining properties will still be set.
        """
        ordered_keys = (1, 2, 3)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("obj", "prop_set", "names", "from_obj", "from_props", "vals")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("set_properties() got an unexpected keyword argument")
            keys = ("obj", "prop_set")
            for key in keys:
                if key in kwargs:
                    ka[1] = kwargs[key]
                    break
            keys = ("names", "from_obj", "from_props")
            for key in keys:
                if key in kwargs:
                    ka[2] = kwargs[key]
                    break
            if count == 2:
                return ka
            ka[3] = kwargs.get("vals", None)
            return ka

        if not count in (2, 3):
            raise TypeError("set_properties() got an invalid numer of arguments")

        kargs = get_kwargs()
        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg

        if mInfo.Info.is_type_interface(kargs[1], "com.sun.star.beans.XPropertySet"):
            # set_properties(cls, prop_set: XPropertySet, names: Sequence[str], vals: Sequence[object]) -> None
            # set_properties(cls, prop_set: XPropertySet, from_props: XPropertySet) -> None
            prop_set = cast(XPropertySet, kargs[1])
        else:
            # set_properties(cls, obj: object, names: Sequence[str], vals: Sequence[object]) -> None
            # set_properties(cls, obj: object, from_obj: object) -> None
            prop_set = mLo.Lo.qi(XPropertySet, kargs[1])
            if prop_set is None:
                raise mEx.MissingInterfaceError(XPropertySet)

        if count == 3:
            # set_properties(cls, prop_set: XPropertySet, names: Sequence[str], vals: Sequence[object]) -> None
            # set_properties(cls, obj: object, names: Sequence[str], vals: Sequence[object]) -> None
            cls._set_properties_by_vals(prop_set=prop_set, names=kargs[2], vals=kargs[3])
            return

        elif count == 2:
            if mInfo.Info.is_type_interface(kargs[2], "com.sun.star.beans.XPropertySet"):
                # set_properties(cls, prop_set: XPropertySet, from_props: XPropertySet) -> None
                from_props = cast(XPropertySet, kargs[2])
            else:
                # set_properties(cls, obj: object, from_obj: object) -> None
                from_props = mLo.Lo.qi(XPropertySet, kargs[1], True)
            cls._set_properties_from_props(prop_set=prop_set, from_props=from_props)

    @classmethod
    def _set_properties_by_vals(cls, prop_set: XPropertySet, names: Sequence[str], vals: Sequence[object]) -> None:
        errs = []
        for i, name in enumerate(names):
            try:
                prop_set.setPropertyValue(name, vals[i])
            except PropertyVetoException as e:
                errs.append(mEx.PropertySetError(f'Could not set readonly-property "{name}"', e))
            except UnknownPropertyException as e:
                errs.append(mEx.PropertyNotFoundError(name, e))
            except Exception as e:
                errs.append(Exception(f'Could not set property "{name}"', e))
        if len(errs) > 0:
            raise mEx.MultiError(errs)

    @classmethod
    def _set_properties_from_props(cls, prop_set: XPropertySet, from_props: XPropertySet) -> None:
        errs = []
        if prop_set is None:
            print(f"Property set is null; cannot set properties")
            return
        if from_props is None:
            print("Source property set is null; cannot set properties")
            return
        nms = cls.get_prop_names(from_props)
        for itm in nms:
            try:
                prop_set.setPropertyValue(itm, cls.get_property(from_props, itm))
            except PropertyVetoException as e:
                errs.append(mEx.PropertySetError(f'Could not set readonly-property "{itm}"', e))
            except UnknownPropertyException as e:
                errs.append(mEx.PropertyNotFoundError(itm, e))
            except Exception as e:
                errs.append(Exception(f'Could not set property "{itm}"', e))
        if len(errs) > 0:
            raise mEx.MultiError(errs)

    # endregion set_properties()

    @staticmethod
    def _set_by_attribute(obj: object, name: str, value: Any) -> bool:
        # there is a bug in LibreOffice that in some cases getting XPrpertySet
        # returns the interface but it is missing setPropertyValue and or getPropertySetInfo
        # this is the case with XChartDocument.getPageBackground() and XChartDocument.getFirstDiagram().getWall()
        # See Chart2.set_background_colors()
        #
        # the same issue (ct is XChartType) with:
        #   white_day_ps = mLo.Lo.qi(XPropertySet, mProps.Props.get(ct, "WhiteDay"), True)
        #   black_day_ps = mLo.Lo.qi(XPropertySet, mProps.Props.get(ct, "BlackDay"), True)
        # see Chart2.color_stock_bars()
        # Even though the interface is misssing setPropertyValue and or getPropertySetInfo the
        # properties are still there.
        # That is why this method. It is a fallback when bug is found to set value by attribute.
        # see Props.set()

        if not hasattr(obj, name):
            return False
        try:
            setattr(obj, name, value)
            return True
        except AttributeError:
            pass
        return False

    @staticmethod
    def _get_by_attribute(obj: object, name: str) -> Tuple[bool, Any]:
        # there is a bug in LibreOffice that in some cases getting XPrpertySet
        # returns the interface but it is missing setPropertyValue and or getPropertySetInfo
        # this is the case with XChartDocument.getPageBackground() and XChartDocument.getFirstDiagram().getWall()
        # See Chart2.set_background_colors()
        #
        # the same issue (ct is XChartType) with:
        #   white_day_ps = mLo.Lo.qi(XPropertySet, mProps.Props.get(ct, "WhiteDay"), True)
        #   black_day_ps = mLo.Lo.qi(XPropertySet, mProps.Props.get(ct, "BlackDay"), True)
        # see Chart2.color_stock_bars()
        # Even though the interface is misssing setPropertyValue and or getPropertySetInfo the
        # properties are still there.
        # That is why this method. It is a fallback when bug is found to get value by attribute.
        # see Props.get()
        try:
            return (True, getattr(obj, name))
        except AttributeError:
            pass
        return (False, None)

    @classmethod
    def set(cls, obj: object, **kwargs) -> None:
        """
        Set one or more properties.

        Args:
            obj (object): object to set properties for. Must support ``XPropertySet``
            **kwargs: Variable length Key value pairs used to set properties.

        Raises:
            MissingInterfaceError: if obj does not implement XPropertySet interface
            MultiError: If unable to set a property

        Returns:
            None:

        Note:
            If ``MultiError`` occurs only the properties that raised an error is part of the error object.
            The remaining properties will still be set.
        """
        if len(kwargs) == 0:
            return
        if mInfo.Info.is_type_interface(obj, "com.sun.star.beans.XPropertySet"):
            ps = cast(XPropertySet, obj)
        else:
            ps = mLo.Lo.qi(XPropertySet, obj, True)
        errs = []
        for key, value in kwargs.items():
            try:
                ps.setPropertyValue(key, value)
            except AttributeError as e:
                # handle a LibreOffice bug
                if not cls._set_by_attribute(obj, key, value):
                    errs.append(
                        mEx.UnKnownError(
                            f'Something went wrong. Could not find setPropertyValue attribute on property set. Tried setting "{key}" manually but failed'
                        )
                    )
                    errs.append(e)
            except PropertyVetoException as e:
                errs.append(mEx.PropertySetError(f'Could not set readonly-property "{key}"', e))
            except UnknownPropertyException as e:
                errs.append(mEx.PropertyNotFoundError(key, e))
            except Exception as e:
                errs.append(Exception(f'Could not set property "{key}"', e))
        if len(errs) > 0:
            raise mEx.MultiError(errs)

    @classmethod
    def get(cls, obj: object, name: str) -> Any:
        """
        Gets a property value from an object.
        ``obj`` must support ``XPropertySet`` interface

        Args:
            obj (object): Object to get property from.
            name (str): Property Name.

        Raises:
            PropertyNotFoundError: If Property is not found
            PropertyError: If any other error occurs.

        Returns:
            Any: Property value
        """
        try:
            if mInfo.Info.is_type_interface(obj, "com.sun.star.beans.XPropertySet"):
                ps = cast(XPropertySet, obj)
            else:
                ps = mLo.Lo.qi(XPropertySet, obj, True)
            try:
                val = ps.getPropertyValue(name)
            except AttributeError as e:
                # handle a LibreOffice bug
                success, val = cls._get_by_attribute(ps, name)
                if success:
                    return val
                else:
                    raise mEx.UnKnownError(
                        f'Something went wrong. Could not find getPropertyValue attribute on property set. Tried getting "{name}" manually but failed.'
                    )
            # it is perfeclty fine for a property to have a value of None
            return val
        except UnknownPropertyException:
            # the property name is not in the property set
            raise mEx.PropertyNotFoundError(name)
        except Exception as e:
            raise mEx.PropertyError(f'Error getting property: "{name}"') from e

    # endregion ---------------- set properties -----------------------

    # region ------------------- get properties ------------------------

    # region    get_property()
    @overload
    @classmethod
    def get_property(cls, obj: object, name: str) -> object:
        """
        Gets a property value from property set

        Args:
            obj (object): object that implements XPropertySet interface
            xprops (XPropertySet): property set
            name (str): property name

        Raises:
            PropertyNotFoundError: If unable to get property
            MissingInterfaceError: if obj does not implement XPropertySet interface

        Returns:
            object: property value
        """
        ...

    @overload
    @classmethod
    def get_property(cls, prop_set: XPropertySet, name: str) -> object:
        """
        Gets a property value from property set

        Args:
            obj (object): object that implements XPropertySet interface
            xprops (XPropertySet): property set
            name (str): property name

        Raises:
            PropertyNotFoundError: If unable to get property

        Returns:
            object: property value
        """
        ...

    @classmethod
    def get_property(cls, *args, **kwargs) -> object:
        """
        Gets a property value from property set

        Args:
            obj (object): object that implements XPropertySet interface
            xprops (XPropertySet): property set
            name (str): property name

        Raises:
            PropertyNotFoundError: If Property is not found
            PropertyError: If any other error occurs.

        Returns:
            object: property value
        """
        ordered_keys = (1, 2)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("obj", "prop_set", "name")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("get_property() got an unexpected keyword argument")
            keys = ("obj", "prop_set")
            for key in keys:
                if key in kwargs:
                    ka[1] = kwargs[key]
                    break
            ka[2] = kwargs.get("name", None)
            return ka

        if count != 2:
            raise TypeError("get_property() got an invalid numer of arguments")

        kargs = get_kwargs()
        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg

        return cls.get(kargs[1], kargs[2])

        # if mInfo.Info.is_type_interface(kargs[1], XPropertySet.__pyunointerface__):
        #     prop_set = cast(XPropertySet, kargs[1])
        # else:
        #     prop_set = mLo.Lo.qi(XPropertySet, kargs[1])
        #     if prop_set is None:
        #         raise mEx.MissingInterfaceError(XPropertySet)
        # name = kargs[2]
        # try:
        #     try:
        #         return prop_set.getPropertyValue(name)
        #     except RuntimeException as e:
        #         mEx.PropertyError(name, f"Could not get runtime property '{name}': {e}")
        # except Exception as e:
        #     raise mEx.PropertyNotFoundError(prop_name=name) from e

    # endregion    get_property()

    @staticmethod
    def get_properties(obj: object) -> Tuple[Property, ...]:
        """
        Get properties

        Args:
            obj (object): Object that implements XPropertySet interface

        Raises:
            MissingInterfaceError: if obj dos not implement XPropertySet

        Returns:
            Tuple[Property, ...]: Properties of obj
        """
        prop_set = mLo.Lo.qi(XPropertySet, obj)
        if prop_set is None:
            raise mEx.MissingInterfaceError(XPropertySet)
        props = list(prop_set.getPropertySetInfo().getProperties())
        props.sort(key=lambda prop: prop.Name)
        return tuple(props)

    @staticmethod
    def get_prop_names(obj: object) -> Tuple[str, ...]:
        """
        Gets property names

        Args:
            obj (object): Object that implements XPropertySet interface

        Raises:
            MissingInterfaceError: if obj dos not implement XPropertySet

        Returns:
            Tuple[str, ...]: Property names of obj
        """
        prop_set = mLo.Lo.qi(XPropertySet, obj)
        if prop_set is None:
            raise mEx.MissingInterfaceError(XPropertySet)
        props = prop_set.getPropertySetInfo().getProperties()
        nms = []
        for prop in props:
            nms.append(prop.Name)
        return tuple(nms)

    @staticmethod
    def get_value(name: str, props: Iterable[PropertyValue]) -> object:
        """
        Get a property value from properties

        Args:
            name (str): property name
            props (Iterable[PropertyValue]): Properties to search

        Raises:
            PropertyError: If property name is not found.

        Returns:
            object: Property Value
        """
        for prop in props:
            if prop.Name == name:
                return prop.Value
        raise mEx.PropertyNotFoundError(
            name,
        )

    # endregion ---------------- get properties ------------------------

    # region ------------------- show properties array -----------------

    @classmethod
    def show_indexed_props(cls, title: str, obj: object) -> None:
        """
        Prints objects properties to console

        Args:
            title (str): Title to print
            obj (object): object that implements XIndexAccess

        Example:
            .. code-block:: python

                >>> from ooodev.office.calc import Calc
                >>> from ooodev.utils.props import Props
                >>> from ooodev.utils.lo import Lo
                >>> loader = Lo.load_office()
                >>> doc = Calc.create_doc(loader)
                >>> sheets = Calc.get_sheets(doc)
                >>> Props.show_indexed_props("Sheets Property", sheets)
                Indexed Properties for 'Sheets Property':
                No. of elements: 1
                Elem 0 Properties
                  AbsoluteName: $Sheet1.$A$1:$AMJ$1048576
                  AsianVerticalMode: False
                  AutomaticPrintArea: True
                  BorderColor: None
                  BottomBorder: (com.sun.star.table.BorderLine2){ (com.sun.star.table.BorderLine){ Color = (long)0x0, InnerLineWidth = (short)0x0, OuterLineWidth = (short)0x0, LineDistance = (short)0x0 }, LineStyle = (short)0x0, LineWidth = (unsigned long)0x0 }
                  BottomBorder2: (com.sun.star.table.BorderLine2){ (com.sun.star.table.BorderLine){ Color = (long)0x0, InnerLineWidth = (short)0x0, OuterLineWidth = (short)0x0, LineDistance = (short)0x0 }, LineStyle = (short)0x0, LineWidth = (unsigned long)0x0 }
                  CellBackColor: -1
                  CellProtection: (com.sun.star.util.CellProtection){ IsLocked = (boolean)true, IsFormulaHidden = (boolean)false, IsHidden = (boolean)false, IsPrintHidden = (boolean)false }
                  CellStyle: Default
                  CharColor: -1
                  ...
        """
        print(f"Indexed Properties for '{title}':")
        in_acc = mLo.Lo.qi(XIndexAccess, obj)
        if in_acc is None:
            print("Could not convert object to an IndexAccess container")
            return

        num_elems = in_acc.getCount()
        print(f"No. of elements: {num_elems}")
        if num_elems == 0:
            return
        for i in range(num_elems):
            try:
                # PropertyValue[] props = Lo.qi(PropertyValue[].class, inAcc.getByIndex(i));
                # above line is original java code.
                # perhapsh thre is a way to alos include PropertyValue[] queryInterface
                props = mLo.Lo.qi(XPropertySet, in_acc.getByIndex(i))
                if props is None:
                    return

                cls.show_props(f"Elem {i}", props)
                print("----")
            except Exception as e:
                print(f"Could not get elem {i}: {e}")

    @classmethod
    def prop_value_to_string(cls, val: object) -> str:
        """
        Gets property values a a string

        Args:
            val (object): Values such as a iterable of iterable or
                object that implements XPropertySet or
                a string

        Returns:
            str: A string representing properties

        Example:
             .. code-block:: python

                >>> from ooodev.office.calc import Calc
                >>> from ooodev.utils.props import Props
                >>> from ooodev.utils.lo import Lo
                >>> loader = Lo.load_office()
                >>> doc = Calc.create_doc(loader)
                >>> sheet = Calc.get_sheet(doc=doc, index=0)
                >>> prop_str = Props.prop_value_to_string(sheet)
                >>> print(prop_str)
                [
                    AbsoluteName = $Sheet1.$A$1:$AMJ$1048576
                    AsianVerticalMode = False
                    AutomaticPrintArea = True
                    BorderColor = None
                    BottomBorder = (com.sun.star.table.BorderLine2){ (com.sun.star.table.BorderLine){ Color = (long)0x0, InnerLineWidth = (short)0x0, OuterLineWidth = (short)0x0, LineDistance = (short)0x0 }, LineStyle = (short)0x0, LineWidth = (unsigned long)0x0 }
                    BottomBorder2 = (com.sun.star.table.BorderLine2){ (com.sun.star.table.BorderLine){ Color = (long)0x0, InnerLineWidth = (short)0x0, OuterLineWidth = (short)0x0, LineDistance = (short)0x0 }, LineStyle = (short)0x0, LineWidth = (unsigned long)0x0 }
                    CellBackColor = -1
                    CellProtection = (com.sun.star.util.CellProtection){ IsLocked = (boolean)true, IsFormulaHidden = (boolean)false, IsHidden = (boolean)false, IsPrintHidden = (boolean)false }
                    CellStyle = Default
                    CharColor = -1
                    ...
                ]
        """

        def get_pv_str(vals) -> str:
            lines = []
            for p in vals:
                try:
                    # value = cls.get_property()
                    lines.append(f"{p.Name} = {p.Value}")
                except AttributeError:
                    continue
            return "[\n    " + "\n    ".join(lines) + "\n]"

        def get_property_set_str(prop_set, props) -> str:
            lines = []
            for p in props:
                value = cls.get(prop_set, p.Name)
                lines.append(f"{p.Name} = {value}")
            return "[\n    " + "\n    ".join(lines) + "\n]"

        if val is None:
            return ""
        if isinstance(val, str):
            return val
        try:
            _ = iter(val)
        except TypeError:
            # not iterable
            is_iter = False
        else:
            # iterable
            is_iter = True
        if is_iter:

            try:
                if isinstance(val[0], str):
                    # assume Iterable[str]
                    return ", ".join(val)
            except Exception:
                pass

            # assume Iterable[PropertyValue]
            return get_pv_str(val)
        else:
            xprops = mLo.Lo.qi(XPropertySet, val)
            if xprops is not None:
                lst = list(cls.props_set_to_tuple(xprops))
                lst.sort(key=lambda prop: prop.Name)
                return get_property_set_str(xprops, lst)
            return str(val)

    @classmethod
    def show_values(cls, name: str, props: Iterable[PropertyValue]) -> None:
        """
        Prints property to console

        Args:
            name (str): Property Name
            props (Iterable[PropertyValue]): Properties
        """
        for prop in props:
            if prop.Name == name:
                print(f"{prop.Name}: {cls.prop_value_to_string(prop.Value)}")
                return
        print(f"{name} not found")

    # endregion ---------------- show properties array -----------------

    # region ------------------- show properties of an Object ----------

    @classmethod
    def show_obj_props(cls, prop_kind: str, obj: object) -> None:
        """
        Prints properties for an object to the console

        Args:
            prop_kind (str): The kind of properties that is displayed in console
            obj (object): object the implements XPropertySet interface.
        """
        prop_set = mLo.Lo.qi(XPropertySet, obj)
        if prop_set is None:
            print(f"no {prop_kind} properties found")
            return
        cls.show_props(prop_kind=prop_kind, props_set=prop_set)

    # region    show_props()
    @overload
    @classmethod
    def show_props(cls, title: str, props: Sequence[PropertyValue]) -> None:
        """
        Prints properties to console

        Args:
            title (str): Title to use that is displayed in console
            props (Sequence[PropertyValue]): Properties to print
        """
        ...

    @overload
    @classmethod
    def show_props(cls, prop_kind: str, props_set: XPropertySet) -> None:
        """
        Prints properties to console

        Args:
            prop_kind (str): The kind of properties that is displayed in console
            props_set (XPropertySet): object the implements XPropertySet interface.
        """
        ...

    @classmethod
    def show_props(cls, *args, **kwargs) -> None:
        """
        Prints properties to console

        Args:
            title (str): Title to use that is displayed in console
            props (Sequence[PropertyValue]): Properties to print
            prop_kind (str): The kind of properties that is displayed in console
            props_set (XPropertySet): object the implements XPropertySet interface.
        """
        ordered_keys = (1, 2)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("title", "props", "prop_kind", "props_set")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("show_props() got an unexpected keyword argument")
            keys = ("title", "prop_kind")
            for key in keys:
                if key in kwargs:
                    ka[1] = kwargs[key]
                    break
            keys = ("props", "props_set")
            for key in keys:
                if key in kwargs:
                    ka[2] = kwargs[key]
                    break
            return ka

        if count != 2:
            raise TypeError("show_props() got an invalid numer of arguments")

        kargs = get_kwargs()

        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg

        if mInfo.Info.is_type_interface(kargs[2], "com.sun.star.beans.XPropertySet"):
            # def show_props(prop_kind: str, props_set: XPropertySet)
            return cls._show_props_str_xpropertyset(prop_kind=kargs[1], props_set=kargs[2])
        else:
            # def show_props(title: str, props: Sequence[PropertyValue])
            return cls._show_props_str_props(title=kargs[1], props=kargs[2])

    @classmethod
    def _show_props_str_props(cls, title: str, props: Sequence[PropertyValue]) -> None:
        print(f"Properties for '{title}':")
        if props is None:
            print("  none found")
            return
        lst = list(props)
        lst.sort(key=lambda prop: prop.Name)
        for prop in lst:

            print(f"  {prop.Name}: {cls.prop_value_to_string(prop.Value)}")
        print()

    @classmethod
    def _show_props_str_xpropertyset(cls, prop_kind: str, props_set: XPropertySet) -> None:
        props = cls.props_set_to_tuple(props_set)
        if props is None:
            print(f"No {prop_kind} properties found")
            return
        lst = list(props)
        lst.sort(key=lambda prop: prop.Name)
        print(f"{prop_kind} Properties")
        for prop in lst:
            prop_value = cls.get_property(props_set, prop.Name)
            print(f"  {prop.Name}: {prop_value}")
        print()

    # endregion  show_props()

    @staticmethod
    def props_set_to_tuple(xprops: XPropertySet) -> Tuple[Property, ...]:
        """
        Converts Property Set to Tuple of Property

        Args:
            xprops (XPropertySet): Property set

        Returns:
            Tuple[Property, ...]: Tuple of Property
        """
        if xprops is None:
            return ()
        xprops_info = xprops.getPropertySetInfo()
        return xprops_info.getProperties()

    props_set_to_array = props_set_to_tuple

    @staticmethod
    def show_property(p: Property) -> str:
        """
        Gets a property Name and type as string.

        Args:
            p (Property): Property to print

        Returns:
            str: Property inf format of 'Name: TypeName'
        """
        # p.Type is uno.Type
        return f"{p.Name}: {p.Type.typeName}"

    # endregion ---------------- show properties of an Object ----------

    # region ------------------- others --------------------------------
    @staticmethod
    def has_property(obj: object, name: str) -> bool:
        """
        Gets if a object contains a property matching name

        Args:
            obj (object): An object that implements XPropertySet
            name (str): Property Name

        Returns:
            bool: True if obj contains Property that matches name; Otherwise, False
        """
        prop_set = mLo.Lo.qi(XPropertySet, obj)
        if prop_set is None:
            return False
        return prop_set.getPropertySetInfo().hasPropertyByName(name)

    @classmethod
    def show_doc_type_props(cls, type: str) -> None:
        """
        Prints doc type info to console

        Args:
            type (str): doc type
        """
        if type is None:
            print("type is None")
            return
        xtype_detect = mLo.Lo.create_instance_mcf(XTypeDetection, "com.sun.star.document.TypeDetection")
        if xtype_detect is None:
            print("No type detector reference")
            return

        xname_access = mLo.Lo.qi(XNameAccess, xtype_detect)
        try:
            props = xname_access.getByName(type)
            cls.show_props(type, props)
        except Exception:
            print(f"No properties for '{type}'")

    @staticmethod
    def get_bound_props(props_set: XMultiPropertySet) -> Tuple[str, ...]:
        """
        Gets bound Properties

        Args:
            props_set (XMultiPropertySet): Multi Property Set

        Returns:
            Tuple[str, ...]: Bound names
        """
        props = props_set.getPropertySetInfo().getProperties()
        names = []
        for p in props:
            is_writable = (p.Attributes & PropertyAttribute.READONLY) == 0
            is_not_null = (p.Attributes & PropertyAttribute.MAYBEVOID) == 0
            is_bound = (p.Attributes & PropertyAttribute.BOUND) == 0
            if is_writable and is_not_null and is_bound:
                names.append(p.Name)
        names_len = len(names)
        if names_len == 0:
            print("No suitable properties were found")
            return ()
        print(f"No. of suitable properties: {names_len}")
        names.sort()
        return tuple(names)

    # endregion ---------------- others --------------------------------



# coding: utf-8
from __future__ import annotations
from typing import Any
from .dispatch_args import AbstractDispacthArgs
from .cancel_event_args import AbstractCancelEventArgs


class AbstractDispatchCancelArgs(AbstractDispacthArgs, AbstractCancelEventArgs):
    __slots__ = ()

    def __init__(self, source: Any, cmd: str, cancel=False) -> None:
        """
        Constructor

        Args:
            source (Any): Event Source
            cmd (str): Event Dispatch Command
            cancel (bool, optional): Cancel value. Defaults to False.
        """
        super().__init__(source=source, cmd=cmd)
        self.cancel = cancel


class DispatchCancelArgs(AbstractDispatchCancelArgs):
    """
    Dispatch Cancel Args
    """

    __slots__ = ("cmd", "source", "_event_name", "event_data", "cancel", "_event_source")

    @staticmethod
    def from_args(args: DispatchCancelArgs) -> DispatchCancelArgs:
        """
        Gets a new instance from existing instance

        Args:
            args (DispatchCancelArgs): Existing Instance

        Returns:
            DispatchCancelArgs: args
        """
        eargs = DispatchCancelArgs(source=args.source, cmd=args.cmd)
        eargs._event_name = args.event_name
        eargs._event_source = args.event_source
        eargs.event_data = args.event_data
        eargs.cancel = args.cancel
        return eargs
